"""Find color card module."""
import cv2
import numpy as np
from plantcv.plantcv import params, outputs, deprecation_warning, fatal_error


def find_color_card(rgb_img, threshold_type='adaptgauss', threshvalue=125, blurry=False, background='dark',
                    record_chip_size="median", label=None):
    """Automatically detect a color card and output info to use in create_color_card_mask function.

    Algorithm written by Brandon Hurr. Updated and implemented into PlantCV by Haley Schuhl.

        Inputs:
    rgb_img          = Input RGB image data containing a color card.
    threshold_type   = Threshold method, either 'normal', 'otsu', or 'adaptgauss', optional (default 'adaptgauss')
    threshvalue      = Thresholding value, optional (default 125)
    blurry           = Bool (default False) if True then image sharpening applied
    background       = Type of image background either 'dark' or 'light' (default 'dark'); if 'light' then histogram
                       expansion applied to better detect edges, but histogram expansion will be hindered if there
                       is a dark background
    record_chip_size = Optional str for choosing chip size measurement to be recorded, either "median",
                       "mean", or None
    label            = Optional label parameter, modifies the variable name of
                       observations recorded (default = pcv.params.sample_label).

    Returns:
    df             = Dataframe containing information about the filtered contours
    start_coord    = Two element tuple of starting coordinates, location of the top left pixel detected
    spacing        = Two element tuple of spacing between centers of chips

    :param rgb_img: numpy.ndarray
    :param threshold_type: str
    :param threshvalue: int
    :param blurry: bool
    :param background: str
    :param record_chip_size: str
    :param label: str
    :return df: pandas.core.frame.DataFrame
    :return start_coord: tuple
    :return spacing: tuple
    """
    deprecation_warning(
        "This function is deprecated and will be removed in PlantCV v5.0. "
        "Please use plantcv.transform.detect_color_card instead."
        )
    # Imports
    import skimage
    import pandas as pd
    from scipy.spatial.distance import squareform, pdist

    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    # Get image attributes
    height, width, _ = rgb_img.shape
    total_pix = float(height * width)

    # Minimum and maximum square size based upon 12 MP image
    min_area = 1000. / 12000000. * total_pix
    max_area = 8000000. / 12000000. * total_pix

    # Create gray image for further processing
    gray_img = cv2.cvtColor(rgb_img, cv2.COLOR_BGR2GRAY)

    # Laplacian Fourier Transform detection of blurriness
    blurfactor = cv2.Laplacian(gray_img, cv2.CV_64F).var()

    # If image is blurry then try to deblur using kernel
    if blurry:
        # from https://www.packtpub.com/mapt/book/Application+Development/9781785283932/2/ch02lvl1sec22/Sharpening
        kernel = np.array([[-1, -1, -1, -1, -1],
                           [-1, 2, 2, 2, -1],
                           [-1, 2, 8, 2, -1],
                           [-1, 2, 2, 2, -1],
                           [-1, -1, -1, -1, -1]]) / 8.0
        # Store result back out for further processing
        gray_img = cv2.filter2D(gray_img, -1, kernel)

    # In darker samples, the expansion of the histogram hinders finding the squares due to problems with the otsu
    # thresholding. If your image has a bright background then apply
    if background == 'light':
        clahe = cv2.createCLAHE(clipLimit=3.25, tileGridSize=(4, 4))
        # apply CLAHE histogram expansion to find squares better with canny edge detection
        gray_img = clahe.apply(gray_img)
    elif background != 'dark':
        fatal_error('Background parameter ' + str(background) + ' is not "light" or "dark"!')

    # Thresholding
    if threshold_type.upper() == "OTSU":
        # Blur slightly so defects on card squares and background patterns are less likely to be picked up
        gaussian = cv2.GaussianBlur(gray_img, (5, 5), 0)
        _, threshold = cv2.threshold(gaussian, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    elif threshold_type.upper() == "NORMAL":
        # Blur slightly so defects on card squares and background patterns are less likely to be picked up
        gaussian = cv2.GaussianBlur(gray_img, (5, 5), 0)
        _, threshold = cv2.threshold(gaussian, threshvalue, 255, cv2.THRESH_BINARY)
    elif threshold_type.upper() == "ADAPTGAUSS":
        # Blur slightly so defects on card squares and background patterns are less likely to be picked up
        gaussian = cv2.GaussianBlur(gray_img, (11, 11), 0)
        threshold = cv2.adaptiveThreshold(gaussian, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                          cv2.THRESH_BINARY_INV, 51, 2)
    else:
        fatal_error('Input threshold_type=' + str(threshold_type) + ' but should be "otsu", "normal", or "adaptgauss"!')

    # Apply automatic Canny edge detection using the computed median
    canny_edges = skimage.feature.canny(threshold)
    canny_edges.dtype = 'uint8'

    # Compute contours to find the squares of the card
    contours, hierarchy = cv2.findContours(canny_edges, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[-2:]
    # Variable of which contour is which
    mindex = []
    # Variable to store moments
    mu = []
    # Variable to x,y coordinates in tuples
    mc = []
    # Variable to x coordinate as integer
    mx = []
    # Variable to y coordinate as integer
    my = []
    # Variable to store area
    marea = []
    # Variable to store whether something is a square (1) or not (0)
    msquare = []
    # Variable to store square approximation coordinates
    msquarecoords = []
    # Variable to store child hierarchy element
    mchild = []
    # Fitted rectangle height
    mheight = []
    # Fitted rectangle width
    mwidth = []
    # Ratio of height/width
    mwhratio = []

    # Extract moments from contour image
    for x in range(0, len(contours)):
        mu.append(cv2.moments(contours[x]))
        marea.append(cv2.contourArea(contours[x]))
        mchild.append(int(hierarchy[0][x][2]))
        mindex.append(x)

    # Cycle through moment data and compute location for each moment
    for m in mu:
        if m['m00'] != 0:  # This is the area term for a moment
            mc.append((int(m['m10'] / m['m00']), int(m['m01']) / m['m00']))
            mx.append(int(m['m10'] / m['m00']))
            my.append(int(m['m01'] / m['m00']))
        else:
            mc.append((0, 0))
            mx.append((0))
            my.append((0))

    # Loop over our contours and extract data about them
    for index, c in enumerate(contours):
        # Area isn't 0, but greater than min-area and less than max-area
        if marea[index] != 0 and min_area < marea[index] < max_area:
            peri = cv2.arcLength(c, True)
            approx = cv2.approxPolyDP(c, 0.1 * peri, True)
            _, wh, _ = cv2.minAreaRect(c)  # Rotated rectangle
            mwidth.append(wh[0])
            mheight.append(wh[1])
            # In different versions of OpenCV, width and height can be listed in a different order
            # To normalize the ratio we sort them and take the ratio of the longest / shortest
            wh_sorted = list(wh)
            wh_sorted.sort()
            mwhratio.append(wh_sorted[1] / wh_sorted[0])
            msquare.append(len(approx))
            # If the approx contour has 4 points then we can assume we have 4-sided objects
            if len(approx) in (4, 5):
                msquarecoords.append(approx)
            else:  # It's not square
                msquarecoords.append(0)
        else:  # Contour has area of 0, not interesting
            msquare.append(0)
            msquarecoords.append(0)
            mwidth.append(0)
            mheight.append(0)
            mwhratio.append(0)

    # Make a pandas df from data for filtering out junk
    all_contours = {'index': mindex, 'x': mx, 'y': my, 'width': mwidth, 'height': mheight, 'res_ratio': mwhratio,
                    'area': marea, 'square': msquare, 'child': mchild}
    df = pd.DataFrame(all_contours)

    # Add calculated blur factor to output
    df['blurriness'] = blurfactor

    # Filter df for attributes that would isolate squares of reasonable size
    df = df[(df['area'] > min_area) & (df['area'] < max_area) & (df['child'] != -1) &
            (df['square'].isin([4, 5])) & (df['res_ratio'] < 1.2) & (df['res_ratio'] > 0.85)]

    # Filter nested squares from dataframe, was having issues with median being towards smaller nested squares
    df = df[~(df['index'].isin(df['index'] + 1))]

    # Count up squares that are within a given radius, more squares = more likelihood of them being the card
    # Median width of square time 2.5 gives proximity radius for searching for similar squares
    median_sq_width_px = df["width"].median()

    # Squares that are within 6 widths of the current square
    pixeldist = median_sq_width_px * 6
    # Computes euclidean distance matrix for the x and y contour centroids
    distmatrix = pd.DataFrame(squareform(pdist(df[['x', 'y']])))
    # Add up distances that are less than  ones have distance less than pixeldist pixels
    distmatrixflat = distmatrix.apply(lambda dist: dist[dist <= pixeldist].count() - 1, axis=1)

    # Append distprox summary to dataframe
    df = df.assign(distprox=distmatrixflat.values)

    # Compute how similar in area the squares are. lots of similar values indicates card isolate area measurements
    filtered_area = df['area']
    # Create empty matrix for storing comparisons
    sizecomp = np.zeros((len(filtered_area), len(filtered_area)))
    # Double loop through all areas to compare to each other
    for p in range(0, len(filtered_area)):
        for o in range(0, len(filtered_area)):
            big = max(filtered_area.iloc[p], filtered_area.iloc[o])
            small = min(filtered_area.iloc[p], filtered_area.iloc[o])
            pct = 100. * (small / big)
            sizecomp[p][o] = pct

    # How many comparisons given 90% square similarity
    sizematrix = pd.DataFrame(sizecomp).apply(lambda sim: sim[sim >= 90].count() - 1, axis=1)

    # Append sizeprox summary to dataframe
    df = df.assign(sizeprox=sizematrix.values)

    # Reorder dataframe for better printing
    df = df[['index', 'x', 'y', 'width', 'height', 'res_ratio', 'area', 'square', 'child',
             'blurriness', 'distprox', 'sizeprox']]

    # Loosely filter for size and distance (relative size to median)
    minsqwidth = median_sq_width_px * 0.80
    maxsqwidth = median_sq_width_px * 1.2
    df = df[(df['distprox'] >= 5) & (df['sizeprox'] >= 5) & (df['width'] > minsqwidth) &
            (df['width'] < maxsqwidth)]

    # Filter for proximity again to root out stragglers. Find and count up squares that are within given radius,
    # more squares = more likelihood of them being the card. Median width of square time 2.5 gives proximity radius
    # for searching for similar squares
    median_sq_width_px = df["width"].median()

    # Squares that are within 6 widths of the current square
    pixeldist = median_sq_width_px * 5
    # Computes euclidean distance matrix for the x and y contour centroids
    distmatrix = pd.DataFrame(squareform(pdist(df[['x', 'y']])))
    # Add up distances that are less than  ones have distance less than pixeldist pixels
    distmatrixflat = distmatrix.apply(lambda dist: dist[dist <= pixeldist].count() - 1, axis=1)

    # Append distprox summary to dataframe
    df = df.assign(distprox=distmatrixflat.values)

    # Filter results for distance proximity to other squares
    df = df[(df['distprox'] >= 4)]
    # Remove all not numeric values use to_numeric with parameter, errors='coerce' - it replace non numeric to NaNs:
    df['x'] = pd.to_numeric(df['x'], errors='coerce')
    df['y'] = pd.to_numeric(df['y'], errors='coerce')

    # Remove NaN
    df = df.dropna()

    if df['x'].min() is np.nan or df['y'].min() is np.nan:
        fatal_error('No color card found under current parameters')
    else:
        # Extract the starting coordinate
        start_coord = (df['x'].min(), df['y'].min())

        # start_coord = (int(df['X'].min()), int(df['Y'].min()))
        # Calculate the range
        spacingx_short = (df['x'].max() - df['x'].min()) / 3
        spacingy_short = (df['y'].max() - df['y'].min()) / 3
        spacingx_long = (df['x'].max() - df['x'].min()) / 5
        spacingy_long = (df['y'].max() - df['y'].min()) / 5
        # Chip spacing since 4x6 card assumed
        spacing_short = min(spacingx_short, spacingy_short)
        spacing_long = max(spacingx_long, spacingy_long)
        # Smaller spacing measurement might have a chip missing
        spacing = int(max(spacing_short, spacing_long))
        spacing = (spacing, spacing)

    if record_chip_size is not None:
        if record_chip_size.upper() == "MEDIAN":
            chip_size = df.loc[:, "area"].median()
            chip_height = df.loc[:, "height"].median()
            chip_width = df.loc[:, "width"].median()
        elif record_chip_size.upper() == "MEAN":
            chip_size = df.loc[:, "area"].mean()
            chip_height = df.loc[:, "height"].mean()
            chip_width = df.loc[:, "width"].mean()
        else:
            print(str(record_chip_size) + " Is not a valid entry for record_chip_size." +
                  " Must be either 'mean', 'median', or None.")
            chip_size = None
            chip_height = None
            chip_width = None
        # Store into global measurements
        outputs.add_observation(sample=label, variable='color_chip_size', trait='size of color card chips identified',
                                method='plantcv.plantcv.transform.find_color_card', scale='none',
                                datatype=float, value=chip_size, label=str(record_chip_size))
        method = record_chip_size.lower()
        outputs.add_observation(sample=label, variable=f'{method}_color_chip_height',
                                trait=f'{method} height of color card chips identified',
                                method='plantcv.plantcv.transform.find_color_card', scale='none',
                                datatype=float, value=chip_height, label=str(record_chip_size))
        outputs.add_observation(sample=label, variable=f'{method}_color_chip_width',
                                trait=f'{method} size of color card chips identified',
                                method='plantcv.plantcv.transform.find_color_card', scale='none',
                                datatype=float, value=chip_width, label=str(record_chip_size))

    return df, start_coord, spacing



import argparse


class WorkflowInputs:
    """Class for setting workflow inputs in Jupyter."""

    def __init__(self, images: list, names: str, result: str, outdir: str = ".", writeimg: bool = False,
                 debug: str = None, **kwargs):
        """Configure input variables for a PlantCV workflow.

        Keyword arguments:
        images = A list of one or more image files.
        names = A comma-delimited list containing a unique name for each image.
        result = An output JSON or CSV filename.
        outdir = An output directory for saved images (default = '.').
        writeimg = Save output images (default = False).
        debug = Set debug mode to None, 'print', or 'plot' (default = None).
        kwargs = Additional, user-defined workflow inputs.
        """
        self.result = result
        self.outdir = outdir
        self.writeimg = writeimg
        self.debug = debug
        self.__dict__.update(kwargs)
        self.__dict__.update(_name_images(names=names, img_list=images))


def workflow_inputs(*other_args) -> argparse.Namespace:
    """Parse PlantCV workflow command-line options.

    Keyword arguments:
    other_args = additional, user-defined workflow inputs.

    Outputs:
    args = an argparse ArgumentParser object with command-line keyword attributes.

    :param other_args: list
    :return args: argparse.Namespace
    """
    parser = argparse.ArgumentParser(description="PlantCV user workflow.")
    parser.add_argument("images", help="Input image files.", type=str, nargs="+", metavar="IMAGES")
    parser.add_argument("--names", help="A unique name mapped to each input image.", required=True)
    parser.add_argument("--result", help="Output workflow results file.", required=True)
    parser.add_argument("--outdir", help="Output directory for image files.", default=".")
    parser.add_argument("--writeimg", help="Save output images.", default=False, action="store_true")
    parser.add_argument("--debug", help="Turn on debug, prints/plot intermediate images.",
                        choices=[None, "print", "plot"], default=None)
    # Parse additional user-defined workflow inputs
    for arg in other_args:
        parser.add_argument(f"--{arg}", help="Additional, user-defined workflow input.", required=False)
    args = parser.parse_args()

    images = _name_images(names=args.names, img_list=args.images)
    args.__dict__.update(images)
    return args


def _name_images(names: str, img_list: list) -> dict:
    """Pair image names with image file paths.

    Keyword arguments:
    names = a comma-delimited string of unique names.
    img_list = a list of image file paths.

    Outputs:
    images = a dictionary of name-image file path pairings.
    """
    images = {}
    # Enumerate the list of names split from the input comma-delimited string
    for i, name in enumerate(names.split(",")):
        # Pair each name with the corresponding image file path
        images[name.lower()] = img_list[i]
    return images



# Read in a hyperspectral data cube as an array and parse metadata from the header file

import os
import re
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import Spectral_data
from plantcv.plantcv.transform import rescale
from plantcv.plantcv import fatal_error


def _find_closest(spectral_array, target):
    """Find index of a target wavelength band in a hyperspectral data instance.

    Inputs:
        spectral_array = Hyperspectral data instance
        target         = Target wavelength value

    Returns:
        idx            = Index

    :param spectral_array: __main__.Spectral_data
    :param target: float
    :return spectral_array: __main__.Spectral_data
    """
    # Array must be sorted
    idx = spectral_array.searchsorted(target)
    idx = np.clip(idx, 1, len(spectral_array) - 1)
    left = spectral_array[idx - 1]
    right = spectral_array[idx]
    idx -= target - left < right - target
    return idx


def _make_pseudo_rgb(spectral_array):
    """Create the best pseudo-rgb image possible from a hyperspectral datacube

    Inputs:
        spectral_array = Hyperspectral data instance

    Returns:
        pseudo_rgb     = Pseudo-rgb image

    :param spectral_array: __main__.Spectral_data
    :return pseudo_rgb: numpy.ndarray
    """
    # Make shorter variable names for data from the spectral class instance object
    array_data = spectral_array.array_data
    default_bands = spectral_array.default_bands
    wl_keys = spectral_array.wavelength_dict.keys()

    if default_bands is not None:
        pseudo_rgb = cv2.merge((array_data[:, :, int(default_bands[0])],
                                array_data[:, :, int(default_bands[1])],
                                array_data[:, :, int(default_bands[2])]))

    else:
        max_wavelength = max(float(i) for i in wl_keys)
        min_wavelength = min(float(i) for i in wl_keys)
        # Check range of available wavelength
        if max_wavelength >= 600 and min_wavelength <= 490:
            id_red = _find_closest(spectral_array=np.array([float(i) for i in wl_keys]), target=630)
            id_green = _find_closest(spectral_array=np.array([float(i) for i in wl_keys]), target=540)
            id_blue = _find_closest(spectral_array=np.array([float(i) for i in wl_keys]), target=480)

            pseudo_rgb = cv2.merge((array_data[:, :, [id_blue]],
                                    array_data[:, :, [id_green]],
                                    array_data[:, :, [id_red]]))
        else:
            # Otherwise take 3 wavelengths, first, middle and last available wavelength
            id_red = int(len(spectral_array.wavelength_dict)) - 1
            id_green = int(id_red / 2)
            pseudo_rgb = cv2.merge((array_data[:, :, [0]],
                                    array_data[:, :, [id_green]],
                                    array_data[:, :, [id_red]]))

    # Gamma correct pseudo_rgb image
    pseudo_rgb = pseudo_rgb ** (1 / 2.2)
    # Scale each of the channels up to 255
    debug = params.debug
    params.debug = None
    pseudo_rgb = cv2.merge((rescale(pseudo_rgb[:, :, 0]),
                            rescale(pseudo_rgb[:, :, 1]),
                            rescale(pseudo_rgb[:, :, 2])))

    # Reset debugging mode
    params.debug = debug

    return pseudo_rgb


def _find_hdr(filename):
    """Find a header file paired with an hyperspectral data file.

    Keyword arguments:
    filename = File path/name of a hyperspectral data file.

    Returns:
    hdrfile = File path/name of hyperspectral header file.

    :param filename: str
    :return hdrfile: str
    """
    # Split the filename into the path and name
    dat_path, dat_filename = os.path.split(filename)
    # Extract the base of the data file name and the extension, if any
    dat_basename, dat_ext = os.path.splitext(dat_filename)
    # Create a regular expression for the header file
    # The data file extension is included optionally
    hdr_regex = re.compile(f"{dat_basename}({dat_ext})?.hdr")
    # List all the files in the data file directory
    for fname in os.listdir(dat_path):
        # If the filename matches, return the header file path
        if hdr_regex.match(fname):
            return os.path.join(dat_path, fname)
    return None


def _parse_envi(headername):
    """Parse a header file and create dictionary of relevant metadata

    Keyword arguments:
    headername      = File path/name of a hyperspectral data file.

    Returns:
    header_dict     = Dictionary of hdr metadata
    wavelength_dict = Dictionary of wavelength metadata

    :param headername: str
    :return header_dict: dict
    :return wavelength_dict: dict

    """
    # Initialize dictionary
    header_dict = {}
    with open(headername, "r") as f:
        # Replace characters for easier parsing
        hdata = f.read()
        hdata = hdata.replace(",\n", ",")
        hdata = hdata.replace("\n,", ",")
        hdata = hdata.replace("{\n", "{")
        hdata = hdata.replace("\n}", "}")
        hdata = hdata.replace(" \n ", "")
        hdata = hdata.replace(" \n", "")
        hdata = hdata.replace(";", "")
    hdata = hdata.split("\n")

    # Loop through and create a dictionary from the header file
    # Try to reformat strings by replacing all " = " with '=' and " : "
    for string in hdata:
        # Remove white space for consistency across header file formats
        string = string.replace(' ', '')
        if '=' in string:
            header_data = string.split("=")
            header_data[0] = header_data[0].lower()
            header_dict.update({header_data[0].rstrip(): header_data[1].rstrip()})
        elif ':' in string:
            header_data = string.split(":")
            header_data[0] = header_data[0].lower()
            header_dict.update({header_data[0].rstrip(): header_data[1].rstrip()})

    # Reformat wavelengths
    header_dict["wavelength"] = header_dict["wavelength"].replace("{", "")
    header_dict["wavelength"] = header_dict["wavelength"].replace("}", "")
    header_dict["wavelength"] = header_dict["wavelength"].replace(" ", "")
    header_dict["wavelength"] = header_dict["wavelength"].split(",")

    # Create dictionary of wavelengths
    wavelength_dict = {}
    for j, wavelength in enumerate(header_dict["wavelength"]):
        wavelength_dict.update({float(wavelength): float(j)})

    # Replace datatype ID number with the numpy datatype
    dtype_dict = {"1": np.uint8, "2": np.int16, "3": np.int32, "4": np.float32, "5": np.float64, "6": np.complex64,
                  "9": np.complex128, "12": np.uint16, "13": np.uint32, "14": np.int64, "15": np.uint64}
    header_dict["datatype"] = dtype_dict[header_dict["datatype"]]

    return header_dict, wavelength_dict


def _parse_arcgis(headername):
    """Parse a header file and create dictionary of relevant metadata

    Keyword arguments:
    headername      = File path/name of a hyperspectral data file.

    Returns:
    header_dict     = Dictionary of hdr metadata
    wavelength_dict = Dictionary of wavelength metadata

    :param headername: str
    :return header_dict: dict
    :return wavelength_dict: dict

    """
    # Initialize dictionary/lists
    header_dict = {"wavelength": []}
    wavelength_dict = {}
    keyword_dict = {"LAYOUT": "interleave", "NROWS": "lines", "NCOLS": "samples", "NBANDS": "bands",
                    "NBITS": "datatype", "WAVELENGTHS": "wavelength"}

    # Read in metadata
    with open(headername, "r") as f:
        hdata = f.read()
    hdata = hdata.split("\n")  # split on line returns

    # Loop through and create a dictionary from the header file
    for string in hdata:
        header_data = string.upper().split(" ")  # split string on white space
        # If there are two elements then it is a keyword and value pair
        if len(header_data) == 2:
            # Only keep the pair if the keyword is in the keyword dictionary
            if header_data[0] in keyword_dict:
                header_dict[keyword_dict[header_data[0]]] = header_data[1]
        # Otherwise if the line has one element it is either the WAVELENGTH or WAVELENGTH_END keyword
        # or a wavelength value
        elif header_data[0] not in ["WAVELENGTHS", "WAVELENGTHS_END", ""]:
            # Append the wavelength value to the wavelength list
            header_dict["wavelength"].append(header_data[0])
    # Build the wavelength dictionary from the list and index values of wavelengths
    for j, wavelength in enumerate(header_dict["wavelength"]):
        wavelength_dict.update({float(wavelength): float(j)})

    dtype_dict = {"8": np.uint8, "16": np.int16, "32": np.int32, "64": np.float32}
    header_dict["datatype"] = dtype_dict[header_dict["datatype"]]

    return header_dict, wavelength_dict


def read_data(filename, mode="ENVI"):
    """Read hyperspectral image data from file.
    Inputs:
    filename          = Name of image file
    mode              = Format of img data (ENVI or ARCGIS, case insensitive)

    Returns:
    spectral_array    = Hyperspectral data instance

    :param filename: str
    :param mode: str
    :return spectral_array: __main__.Spectral_data
    """
    # Remove any file extension and set .hdr filename
    headername = _find_hdr(filename=filename)

    if headername is None:
        fatal_error(f"Unable to find the header file corresponding to {filename}")

    if mode.upper() == "ENVI":
        header_dict, wavelength_dict = _parse_envi(headername=headername)
    elif mode.upper() == "ARCGIS":
        header_dict, wavelength_dict = _parse_arcgis(headername=headername)

    # Read in the data from the file
    raw_data = np.fromfile(filename, header_dict["datatype"], -1)

    # Reshape the raw data into a datacube array
    data_format = {
        # Band Interleaved by Line (BIL)
        "BIL": {
            # Divide the raw data by Y (lines), Z (spectral bands), and X (samples)
            # Then reorder into a cube in Y, X, Z order
            "reshape": (int(header_dict["lines"]), int(header_dict["bands"]), int(header_dict["samples"])),
            "transpose": (0, 2, 1)
        },
        # Band Interleaved by Pixel (BIP)
        "BIP": {
            # Divide the raw data by Y (lines), and X (samples), Z (spectral bands)
            # Then reorder into a cube in Y, X, Z order
            "reshape": (int(header_dict["lines"]), int(header_dict["samples"]), int(header_dict["bands"])),
            "transpose": (0, 1, 2)
        },
        # Band Sequential (BSQ)
        "BSQ": {
            # Divide the raw data by Z (spectral bands), Y (lines), and X (samples)
            # Then reorder into a cube in Y, X, Z order
            "reshape": (int(header_dict["bands"]), int(header_dict["lines"]), int(header_dict["samples"])),
            "transpose": (1, 2, 0)
        }
    }
    interleave_type = header_dict.get("interleave")
    if interleave_type is not None:
        interleave_type = interleave_type.upper()
    if interleave_type not in data_format:
        fatal_error(f"Interleave type {interleave_type} is not supported.")

    # Reshape raw data into a data cube
    array_data = raw_data.reshape(data_format[interleave_type]["reshape"]
                                  ).transpose(data_format[interleave_type]["transpose"])

    # Check for default bands (that get used to make pseudo_rgb image)
    default_bands = None
    if "defaultbands" in header_dict:
        header_dict["defaultbands"] = header_dict["defaultbands"].replace("{", "")
        header_dict["defaultbands"] = header_dict["defaultbands"].replace("}", "")
        default_bands = header_dict["defaultbands"].split(",")

    # Find array min and max values
    max_pixel = float(np.amax(array_data))
    min_pixel = float(np.amin(array_data))

    wavelength_units = header_dict.get("wavelengthunits")
    if wavelength_units is None:
        wavelength_units = "nm"

        # Create an instance of the spectral_data class
    spectral_array = Spectral_data(array_data=array_data,
                                   max_wavelength=float(str(header_dict["wavelength"][-1]).rstrip()),
                                   min_wavelength=float(str(header_dict["wavelength"][0]).rstrip()),
                                   max_value=max_pixel, min_value=min_pixel,
                                   d_type=header_dict["datatype"],
                                   wavelength_dict=wavelength_dict, samples=int(header_dict["samples"]),
                                   lines=int(header_dict["lines"]), interleave=header_dict["interleave"],
                                   wavelength_units=wavelength_units, array_type="datacube",
                                   pseudo_rgb=None, filename=filename, default_bands=default_bands)

    # Make pseudo-rgb image and replace it inside the class instance object
    pseudo_rgb = _make_pseudo_rgb(spectral_array)
    spectral_array.pseudo_rgb = pseudo_rgb

    _debug(visual=pseudo_rgb, filename=os.path.join(params.debug_outdir, str(params.device) + "_pseudo_rgb.png"))

    return spectral_array



"""Analyzes the color properties of objects in an image."""
import os
import cv2
import numpy as np
from scipy import stats
from plantcv.plantcv import fatal_error
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import outputs
from plantcv.plantcv.visualize import histogram
from plantcv.plantcv._helpers import _iterate_analysis


def color(rgb_img, labeled_mask, n_labels=1, colorspaces="hsv", label=None):
    """A function that analyzes the color of objects and outputs data.

    Inputs:
    rgb_img          = RGB image data.
    labeled_mask     = Labeled mask of objects (32-bit).
    n_labels         = Total number expected individual objects (default = 1).
    colorspaces      = 'all', 'rgb', 'lab', or 'hsv' (default = 'hsv').
    label            = Optional label parameter, modifies the variable name of
                       observations recorded (default = pcv.params.sample_label).

    Returns:
    analysis_image   = histogram output

    :param rgb_img: numpy.ndarray
    :param mask: numpy.ndarray
    :param colorspaces: str
    :param label: str
    :return analysis_images: list
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    _ = _iterate_analysis(img=rgb_img, labeled_mask=labeled_mask, n_labels=n_labels, label=label, function=_analyze_color,
                          **{"colorspaces": colorspaces})
    hue_chart = outputs.plot_dists(variable="hue_frequencies")
    _debug(visual=hue_chart, filename=os.path.join(params.debug_outdir, str(params.device) + '_hue_hist.png'))
    return hue_chart


def _analyze_color(img, mask, colorspaces="hsv", label=None):
    """Analyze the color properties of an image object
    Inputs:
    img              = RGB image data
    mask             = Binary mask made from selected contours
    colorspaces      = 'all', 'rgb', 'lab', or 'hsv'
    label            = optional label parameter, modifies the variable name of observations recorded

    Returns:
    analysis_image   = histogram output

    :param rgb_img: numpy.ndarray
    :param mask: numpy.ndarray
    :param colorspaces: str
    :param label: str
    :return analysis_images: list
    """
    # Initialize output data
    # Histogram plot types
    hist_types = {"all": ("b", "g", "r", "l", "m", "y", "h", "s", "v"),
                  "rgb": ("b", "g", "r"),
                  "lab": ("l", "m", "y"),
                  "hsv": ("h", "s", "v")}

    if colorspaces.lower() not in hist_types:
        fatal_error(f"Colorspace '{colorspaces}' is not supported, must be be one of the following: "
                    f"{', '.join(map(str, hist_types.keys()))}")

    # Empty histograms
    histograms = {
        "b": {"label": "blue", "graph_color": "blue",
              "hist": [0] * 256},
        "g": {"label": "green", "graph_color": "forestgreen",
              "hist": [0] * 256},
        "r": {"label": "red", "graph_color": "red",
              "hist": [0] * 256},
        "l": {"label": "lightness", "graph_color": "dimgray",
              "hist": [0] * 256},
        "m": {"label": "green-magenta", "graph_color": "magenta",
              "hist": [0] * 256},
        "y": {"label": "blue-yellow", "graph_color": "yellow",
              "hist": [0] * 256},
        "h": {"label": "hue", "graph_color": "blueviolet",
              "hist": [0] * 256},
        "s": {"label": "saturation", "graph_color": "cyan",
              "hist": [0] * 256},
        "v": {"label": "value", "graph_color": "orange",
              "hist": [0] * 256}
    }

    # Undefined defaults
    hue_median = np.nan
    hue_circular_mean = np.nan
    hue_circular_std = np.nan

    # Skip empty masks
    if np.count_nonzero(mask) != 0:
        # Save user debug setting
        debug = params.debug
        if len(np.shape(img)) < 3:
            fatal_error("rgb_img must be an RGB image")

        # Mask the input image
        masked = cv2.bitwise_and(img, img, mask=mask)
        # Extract the blue, green, and red channels
        b, g, r = cv2.split(masked)
        # Convert the BGR image to LAB
        lab = cv2.cvtColor(masked, cv2.COLOR_BGR2LAB)
        # Extract the lightness, green-magenta, and blue-yellow channels
        l, m, y = cv2.split(lab)
        # Convert the BGR image to HSV
        hsv = cv2.cvtColor(masked, cv2.COLOR_BGR2HSV)
        # Extract the hue, saturation, and value channels
        h, s, v = cv2.split(hsv)

        # Color channel dictionary
        channels = {"b": b, "g": g, "r": r, "l": l, "m": m, "y": y, "h": h, "s": s, "v": v}

        # Calculate histogram
        params.debug = None
        for channel in hist_types["all"]:
            histograms[channel]["hist"] = histogram(channels[channel], mask, 256, 0, 255,
                                                    hist_data=True)[1]['proportion of pixels (%)'].tolist()

        # Restore user debug setting
        params.debug = debug

        # Hue values of zero are red but are also the value for pixels where hue is undefined. The hue value of a pixel will
        # be undef. when the color values are saturated. Therefore, hue values of 0 are excluded from the calculations below
        # Calculate the median hue value (median is rescaled from the encoded 0-179 range to the 0-359 degree range)
        hue_median = np.median(h[np.where(h > 0)]) * 2

        # Calculate the circular mean and standard deviation of the encoded hue values
        # The mean and standard-deviation are rescaled from the encoded 0-179 range to the 0-359 degree range
        hue_circular_mean = stats.circmean(h[np.where(h > 0)], high=179, low=0) * 2
        hue_circular_std = stats.circstd(h[np.where(h > 0)], high=179, low=0) * 2

    # Store into global measurements
    # RGB signal values are in an unsigned 8-bit scale of 0-255
    rgb_values = list(range(0, 256))
    # Hue values are in a 0-359 degree scale, every 2 degrees at the midpoint of the interval
    hue_values = [i * 2 + 1 for i in range(0, 180)]
    # Percentage values on a 0-100 scale (lightness, saturation, and value)
    percent_values = [round((i / 255) * 100, 2) for i in range(0, 256)]
    # Diverging values on a -128 to 127 scale (green-magenta and blue-yellow)
    diverging_values = list(range(-128, 128))

    if colorspaces.upper() in ('RGB', 'ALL'):
        outputs.add_observation(sample=label, variable='blue_frequencies', trait='blue frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["b"]["hist"], label=rgb_values)
        outputs.add_observation(sample=label, variable='green_frequencies', trait='green frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["g"]["hist"], label=rgb_values)
        outputs.add_observation(sample=label, variable='red_frequencies', trait='red frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["r"]["hist"], label=rgb_values)

    if colorspaces.upper() in ('LAB', 'ALL'):
        outputs.add_observation(sample=label, variable='lightness_frequencies', trait='lightness frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["l"]["hist"], label=percent_values)
        outputs.add_observation(sample=label, variable='green-magenta_frequencies',
                                trait='green-magenta frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["m"]["hist"], label=diverging_values)
        outputs.add_observation(sample=label, variable='blue-yellow_frequencies', trait='blue-yellow frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["y"]["hist"], label=diverging_values)

    if colorspaces.upper() in ('HSV', 'ALL'):
        outputs.add_observation(sample=label, variable='hue_frequencies', trait='hue frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["h"]["hist"][0:180], label=hue_values)
        outputs.add_observation(sample=label, variable='saturation_frequencies', trait='saturation frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["s"]["hist"], label=percent_values)
        outputs.add_observation(sample=label, variable='value_frequencies', trait='value frequencies',
                                method='plantcv.plantcv.analyze.color', scale='frequency', datatype=list,
                                value=histograms["v"]["hist"], label=percent_values)

    # Always save hue stats
    outputs.add_observation(sample=label, variable='hue_circular_mean', trait='hue circular mean',
                            method='plantcv.plantcv.analyze.color', scale='degrees', datatype=float,
                            value=hue_circular_mean, label='degrees')
    outputs.add_observation(sample=label, variable='hue_circular_std', trait='hue circular standard deviation',
                            method='plantcv.plantcv.analyze.color', scale='degrees', datatype=float,
                            value=hue_circular_std, label='degrees')
    outputs.add_observation(sample=label, variable='hue_median', trait='hue median',
                            method='plantcv.plantcv.analyze.color', scale='degrees', datatype=float,
                            value=hue_median, label='degrees')

    return img



from plantcv.plantcv.analyze.color import color
from plantcv.plantcv.analyze.size import size
from plantcv.plantcv.analyze.bound_horizontal import bound_horizontal
from plantcv.plantcv.analyze.bound_vertical import bound_vertical
from plantcv.plantcv.analyze.grayscale import grayscale
from plantcv.plantcv.analyze.thermal import thermal
from plantcv.plantcv.analyze.spectral_reflectance import spectral_reflectance
from plantcv.plantcv.analyze.spectral_index import spectral_index
from plantcv.plantcv.analyze.yii import yii
from plantcv.plantcv.analyze.npq import npq
from plantcv.plantcv.analyze.distribution import distribution

__all__ = ["color", "bound_horizontal", "bound_vertical", "grayscale", "size", "thermal", "spectral_reflectance",
           "spectral_index", "yii", "npq", "distribution"]



"""PlantCV quick_filter module."""
import os
import cv2
import numpy as np
from skimage.measure import label
from skimage.color import label2rgb
from plantcv.plantcv import params
from plantcv.plantcv.roi import roi2mask
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _logical_operation


def quick_filter(mask, roi, roi_type="partial"):
    """Quickly filter a binary mask using a region of interest.

    Parameters
    ----------
    mask : numpy.ndarray
        Binary mask to filter.
    roi : plantcv.plantcv.classes.Objects
        PlantCV ROI object.
    roi_type : str, optional
        filter method, one of "partial", "cutto", or "within"

    Returns
    -------
    numpy.ndarray
        Filtered binary mask.
    """
    # process for cutto is different, so if roi_type is cutto then use helper function
    if roi_type == "cutto":
        return _quick_cutto(mask, roi)[0]
    # Increment the device counter
    params.device += 1

    # Store debug
    debug = params.debug
    params.debug = None

    # Label objects in the image from 1 to n (labeled mask)
    labels, num = label(label_image=mask, return_num=True)

    # Convert the input ROI to a binary mask (only works on single ROIs)
    roi_mask = roi2mask(img=mask, roi=roi)

    # Convert the labeled mask and ROI mask to float data types
    roi_mask = roi_mask.astype(float)
    labels = labels.astype(float)

    # Set the ROI mask value to 0.5
    roi_mask[np.where(roi_mask == 255)] = 0.5

    # Add the labeled mask and ROI mask together
    summed = roi_mask + labels

    # For each label, if at least one pixel of the object overlaps the ROI
    # set all the label values to the label plus 0.5
    for i in range(1, num + 1):
        # For each label, if at least one pixel of the object overlaps the ROI
        # set all the label values to the label plus 0.5
        if roi_type.upper() == "PARTIAL" and i + 0.5 in summed:
            summed[np.where(summed == i)] = i + 0.5
        # If one pixel of the object falls outside the ROI
        # set all the label values to zero
        elif roi_type.upper() == "WITHIN" and i in summed:
            summed[np.where(labels == i)] = 0
    # Objects that do not overlap the ROI will round to an integer and have
    # the same value before and after rounding.
    # Objects that overlap the ROI will round up/down and will not have the same value
    # Where the values are equal (not overlapping)
    summed[np.where(summed == summed.round())] = 0

    # The summed image now only contains objects that overlap the ROI
    # Subtract 0.5 to remove the ROI mask
    summed = summed - 0.5

    # Round and set the data type back to uint8
    summed = summed.round().astype("uint8")

    # Make sure the mask is binary
    summed[np.where(summed > 0)] = 255

    # Print/plot debug image
    params.debug = debug
    _debug(visual=summed, filename=os.path.join(params.debug_outdir, f"{params.device}_roi_filter.png"), cmap="gray")

    return summed


def _quick_cutto(mask, roi):
    """Quickly filter a binary mask using a region of interest by cutting to each ROI.

    Parameters
    ----------
    mask : numpy.ndarray
        Binary mask to filter.
    roi : plantcv.plantcv.classes.Objects
        PlantCV ROI object.

    Returns
    -------
    numpy.ndarray, numpy.ndarray, int
        Filtered binary mask, labeled mask, number of labels.
    """
    # Increment the device counter
    params.device += 1

    # Store debug
    debug = params.debug
    params.debug = None

    mask_copy = np.copy(mask).astype(np.int32)
    labeled_mask = np.zeros(mask.shape, dtype=np.int32)
    bin_mask = np.copy(labeled_mask)
    num_labels = len(roi.contours)
    for i in range(num_labels):
        # Pixel intensity of (i+1) such that the first object has value
        cv2.drawContours(labeled_mask, roi.contours[i], -1, (i+1), -1)
        cv2.drawContours(bin_mask, roi.contours[i], -1, (255), -1)
    cropped_mask = _logical_operation(mask_copy, bin_mask, "and")
    # Make a labeled mask from the cropped objects
    label_mask_where = np.where(cropped_mask == 255, labeled_mask, 0)

    # Print/plot debug image
    colorful = label2rgb(label_mask_where)
    colorful2 = (255*colorful).astype(np.uint8)
    params.debug = debug
    _debug(visual=colorful2, filename=os.path.join(params.debug_outdir, f"{params.device}_label_colored_mask.png"))

    return cropped_mask.astype(np.uint8), label_mask_where, num_labels



"""Threshold functions."""
import os
import cv2
import math
import numpy as np
from matplotlib import pyplot as plt
from plantcv.plantcv import fatal_error, warn, params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rgb2lab, _rgb2hsv, _rgb2gray, _rgb2cmyk
from skimage.feature import graycomatrix, graycoprops
from scipy.ndimage import generic_filter


# Binary threshold
def binary(gray_img, threshold, object_type="light"):
    """Creates a binary image from a grayscale image based on the threshold value.

    Inputs:
    gray_img     = Grayscale image data
    threshold    = Threshold value (0-255)
    object_type  = "light" or "dark" (default: "light")
                   - If object is lighter than the background then standard thresholding is done
                   - If object is darker than the background then inverse thresholding is done

    Returns:
    bin_img      = Thresholded, binary image

    :param gray_img: numpy.ndarray
    :param threshold: int
    :param object_type: str
    :return bin_img: numpy.ndarray
    """
    # Set the threshold method
    threshold_method = ""
    if object_type.upper() == "LIGHT":
        threshold_method = cv2.THRESH_BINARY
    elif object_type.upper() == "DARK":
        threshold_method = cv2.THRESH_BINARY_INV
    else:
        fatal_error('Object type ' + str(object_type) + ' is not "light" or "dark"!')

    params.device += 1

    # Threshold the image
    bin_img = _call_threshold(gray_img, threshold, threshold_method, "_binary_threshold_")

    return bin_img


# Gaussian adaptive threshold
def gaussian(gray_img, ksize, offset, object_type="light"):
    """Creates a binary image from a grayscale image based on the Gaussian adaptive threshold method.

    Adaptive thresholds use a threshold value that varies across the image.
    This local threshold depends on the local average, computed in a squared portion of the image of
    ksize by ksize pixels, and on the offset relative to that local average.

    In the Gaussian adaptive threshold, the local average is a weighed average of the pixel values
    in the block, where the weights are a 2D Gaussian centered in the middle.

    Inputs:
    gray_img     = Grayscale image data
    ksize        = Size of the block of pixels used to compute the local average
    offset       = Value substracted from the local average to compute the local threshold.
                    A negative offset sets the local threshold above the local average.
    object_type  = "light" or "dark" (default: "light")
                   - "light" (for objects brighter than the background) sets the pixels above
                        the local threshold to 255 and the pixels below to 0.
                   - "dark" (for objects darker than the background) sets the pixels below the
                        local threshold to 255 and the pixels above to 0.

    Returns:
    bin_img      = Thresholded, binary image

    :param gray_img: numpy.ndarray
    :param ksize: int
    :param offset: float
    :param object_type: str
    :return bin_img: numpy.ndarray
    """
    # Set the threshold method
    threshold_method = ""
    if object_type.upper() == "LIGHT":
        threshold_method = cv2.THRESH_BINARY
    elif object_type.upper() == "DARK":
        threshold_method = cv2.THRESH_BINARY_INV
    else:
        fatal_error('Object type ' + str(object_type) + ' is not "light" or "dark"!')

    params.device += 1

    bin_img = _call_adaptive_threshold(gray_img, ksize, offset, cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                       threshold_method, "_gaussian_threshold_")

    return bin_img


# Mean adaptive threshold
def mean(gray_img, ksize, offset, object_type="light"):
    """Creates a binary image from a grayscale image based on the mean adaptive threshold method.

    Adaptive thresholds use a threshold value that varies across the image.
    This local threshold depends on the local average, computed in a squared portion of the image of
    ksize by ksize pixels, and on the offset relative to that local average.

    In the mean adaptive threshold, the local average is the average of the pixel values in the block.

    Inputs:
    gray_img     = Grayscale image data
    ksize        = Size of the block of pixels used to compute the local average
    offset       = Value substracted from the local average to compute the local threshold.
                    A negative offset sets the local threshold above the local average.
    object_type  = "light" or "dark" (default: "light")
                   - "light" (for objects brighter than the background) sets the pixels above
                        the local threshold to 255 and the pixels below to 0.
                   - "dark" (for objects darker than the background) sets the pixels below the
                        local threshold to 255 and the pixels above to 0.

    Returns:
    bin_img      = Thresholded, binary image

    :param gray_img: numpy.ndarray
    :param ksize: int
    :param offset: float
    :param object_type: str
    :return bin_img: numpy.ndarray
    """
    # Set the threshold method
    threshold_method = ""
    if object_type.upper() == "LIGHT":
        threshold_method = cv2.THRESH_BINARY
    elif object_type.upper() == "DARK":
        threshold_method = cv2.THRESH_BINARY_INV
    else:
        fatal_error('Object type ' + str(object_type) + ' is not "light" or "dark"!')

    params.device += 1

    bin_img = _call_adaptive_threshold(gray_img, ksize, offset, cv2.ADAPTIVE_THRESH_MEAN_C,
                                       threshold_method, "_mean_threshold_")

    return bin_img


# Otsu autothreshold
def otsu(gray_img, object_type="light"):
    """Creates a binary image from a grayscale image using Otsu's thresholding.

    Inputs:
    gray_img     = Grayscale image data
    object_type  = "light" or "dark" (default: "light")
                   - If object is lighter than the background then standard thresholding is done
                   - If object is darker than the background then inverse thresholding is done

    Returns:
    bin_img      = Thresholded, binary image

    :param gray_img: numpy.ndarray
    :param object_type: str
    :return bin_img: numpy.ndarray
    """
    # Set the threshold method
    threshold_method = ""
    if object_type.upper() == "LIGHT":
        threshold_method = cv2.THRESH_BINARY + cv2.THRESH_OTSU
    elif object_type.upper() == "DARK":
        threshold_method = cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU
    else:
        fatal_error('Object type ' + str(object_type) + ' is not "light" or "dark"!')

    params.device += 1

    # Threshold the image
    bin_img = _call_threshold(gray_img, 0, threshold_method, "_otsu_threshold_")

    return bin_img


# Triangle autothreshold
def triangle(gray_img, object_type="light", xstep=1):
    """Creates a binary image from a grayscale image using Zack et al.'s (1977) thresholding.

    Inputs:
    gray_img     = Grayscale image data
    object_type  = "light" or "dark" (default: "light")
                   - If object is lighter than the background then standard thresholding is done
                   - If object is darker than the background then inverse thresholding is done
    xstep        = value to move along x-axis to determine the points from which to calculate distance recommended to
                   start at 1 and change if needed)

    Returns:
    bin_img      = Thresholded, binary image

    :param gray_img: numpy.ndarray
    :param object_type: str
    :param xstep: int
    :return bin_img: numpy.ndarray
    """
    # Calculate automatic threshold value based on triangle algorithm
    hist = cv2.calcHist([gray_img], [0], None, [256], [0, 255])

    # Make histogram one array
    newhist = []
    for item in hist:
        newhist.extend(item)

    # Detect peaks
    show = False
    if params.debug == "plot":
        show = True
    ind = _detect_peaks(newhist, mph=None, mpd=1, show=show)

    # Find point corresponding to highest peak
    # Find intensity value (y) of highest peak
    max_peak_int = max(newhist[i] for i in ind)
    # Find value (x) of highest peak
    max_peak = [i for i, x in enumerate(newhist) if x == max(newhist)]
    # Combine x,y
    max_peak_xy = [max_peak[0], max_peak_int]

    # Find final point at end of long tail
    end_x = len(newhist) - 1
    end_y = newhist[end_x]
    end_xy = [end_x, end_y]

    # Define the known points
    points = [max_peak_xy, end_xy]
    x_coords, y_coords = zip(*points)

    # Get threshold value
    peaks = []
    dists = []

    for i in range(x_coords[0], x_coords[1], xstep):
        distance = (((x_coords[1] - x_coords[0]) * (y_coords[0] - hist[i])) -
                    ((x_coords[0] - i) * (y_coords[1] - y_coords[0]))) / math.sqrt(
            (float(x_coords[1]) - float(x_coords[0])) *
            (float(x_coords[1]) - float(x_coords[0])) +
            ((float(y_coords[1]) - float(y_coords[0])) *
             (float(y_coords[1]) - float(y_coords[0]))))
        peaks.append(i)
        dists.append(distance)
    autothresh = [peaks[x] for x in [i for i, x in enumerate(list(dists)) if x == max(list(dists))]]
    autothreshval = autothresh[0]

    # Set the threshold method
    threshold_method = ""
    if object_type.upper() == "LIGHT":
        threshold_method = cv2.THRESH_BINARY + cv2.THRESH_OTSU
    elif object_type.upper() == "DARK":
        threshold_method = cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU
    else:
        fatal_error('Object type ' + str(object_type) + ' is not "light" or "dark"!')

    params.device += 1

    # Threshold the image
    bin_img = _call_threshold(gray_img, autothreshval, threshold_method, "_triangle_threshold_")

    # Additional figures created by this method, if debug is on
    if params.debug is not None:
        if params.debug == 'print':
            _, ax = plt.subplots()
            ax.plot(hist)
            ax.set(title=f"Threshold value = {autothreshval}")
            ax.axis([0, 256, 0, max(hist)])
            ax.grid(True)
            fig_name_hist = os.path.join(params.debug_outdir,
                                         str(params.device) + '_triangle_thresh_hist_' + str(autothreshval) + ".png")
            # write the figure to current directory
            plt.savefig(fig_name_hist, dpi=params.dpi)
            # close pyplot plotting window
            plt.clf()
        elif params.debug == 'plot':
            print(f"Threshold value = {autothreshval}")
            _, ax = plt.subplots()
            ax.plot(hist)
            ax.axis([0, 256, 0, max(hist)])
            ax.grid(True)
            plt.show()

    return bin_img


def texture(gray_img, ksize, threshold, offset=3, texture_method='dissimilarity', borders='nearest'):
    """Creates a binary image from a grayscale image using skimage texture calculation for thresholding.
    This function is quite slow.

    Inputs:
    gray_img       = Grayscale image data
    ksize          = Kernel size for texture measure calculation
    threshold      = Threshold value (0-255)
    offset         = Distance offsets
    texture_method = Feature of a grey level co-occurrence matrix, either
                     'contrast', 'dissimilarity', 'homogeneity', 'ASM', 'energy',
                     or 'correlation'.For equations of different features see
                     scikit-image.
    borders        = How the array borders are handled, either 'reflect',
                     'constant', 'nearest', 'mirror', or 'wrap'

    Returns:
    bin_img        = Thresholded, binary image

    :param gray_img: numpy.ndarray
    :param ksize: int
    :param threshold: int
    :param offset: int
    :param texture_method: str
    :param borders: str
    :return bin_img: numpy.ndarray
    """
    # Function that calculates the texture of a kernel
    def calc_texture(inputs):
        """Kernel calculate texture function.

        Parameters
        ----------
        inputs : numpy.ndarray
            Input pixel array

        Returns
        -------
        float
            Texture value
        """
        inputs = np.reshape(inputs, newshape=[ksize, ksize])
        inputs = inputs.astype(np.uint8)
        # Greycomatrix takes image, distance offset, angles (in radians), symmetric, and normed
        # http://scikit-image.org/docs/dev/api/skimage.feature.html#skimage.feature.graycomatrix
        glcm = graycomatrix(inputs, [offset], [0], 256, symmetric=True, normed=True)
        diss = graycoprops(glcm, texture_method)[0, 0]
        return diss

    # Make an array the same size as the original image
    output = np.zeros(gray_img.shape, dtype=gray_img.dtype)

    # Apply the texture function over the whole image
    generic_filter(gray_img, calc_texture, size=ksize, output=output, mode=borders)

    # Threshold so higher texture measurements stand out
    bin_img = binary(gray_img=output, threshold=threshold, object_type='light')

    _debug(visual=bin_img, filename=os.path.join(params.debug_outdir, str(params.device) + "_texture_mask.png"))

    return bin_img


def custom_range(img, lower_thresh, upper_thresh, channel='gray'):
    """Creates a thresholded image and mask from an RGB image and threshold values.

    Inputs:
    img      = RGB or grayscale image data
    lower_thresh = List of lower threshold values (0-255)
    upper_thresh = List of upper threshold values (0-255)
    channel      = Color-space channels of interest (RGB, HSV, LAB, or gray)

    Returns:
    mask         = Mask, binary image
    masked_img   = Masked image, keeping the part of image of interest

    :param img: numpy.ndarray
    :param lower_thresh: list
    :param upper_thresh: list
    :param channel: str
    :return mask: numpy.ndarray
    :return masked_img: numpy.ndarray
    """
    if channel.upper() == 'HSV':

        # Check threshold inputs
        _check_threshold_inputs(3, lower_thresh, upper_thresh)

        # Convert the RGB image to HSV colorspace
        hsv_img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)

        # Separate channels
        hue = hsv_img[:, :, 0]
        sat = hsv_img[:, :, 1]
        value = hsv_img[:, :, 2]

        # Make a mask for each channel (Use '*' to avoid linter error about unbalanced value unpacking)
        *out_masks, = _call_inrange(gray_imgs_list=[hue, sat, value],
                                    lower_thresh=lower_thresh, upper_thresh=upper_thresh)

        # Apply the masks to the image
        result = cv2.bitwise_and(img, img, mask=out_masks[0])
        result = cv2.bitwise_and(result, result, mask=out_masks[1])
        masked_img = cv2.bitwise_and(result, result, mask=out_masks[2])

        # Combine masks
        mask = cv2.bitwise_and(out_masks[0], out_masks[1])
        mask = cv2.bitwise_and(mask, out_masks[2])

    elif channel.upper() == 'RGB':

        # Check threshold inputs
        _check_threshold_inputs(3, lower_thresh, upper_thresh)

        # Separate channels (pcv.readimage reads RGB images in as BGR)
        blue = img[:, :, 0]
        green = img[:, :, 1]
        red = img[:, :, 2]

        # Make a mask for each channel (Use '*' to avoid linter error about unbalanced value unpacking)
        *out_masks, = _call_inrange(gray_imgs_list=[blue, green, red],
                                    lower_thresh=lower_thresh, upper_thresh=upper_thresh)

        # Apply the masks to the image
        result = cv2.bitwise_and(img, img, mask=out_masks[0])
        result = cv2.bitwise_and(result, result, mask=out_masks[1])
        masked_img = cv2.bitwise_and(result, result, mask=out_masks[2])

        # Combine masks
        mask = cv2.bitwise_and(out_masks[0], out_masks[1])
        mask = cv2.bitwise_and(mask, out_masks[2])

    elif channel.upper() == 'LAB':

        # Check threshold inputs
        _check_threshold_inputs(3, lower_thresh, upper_thresh)

        # Convert the RGB image to LAB colorspace
        lab_img = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)

        # Separate channels (pcv.readimage reads RGB images in as BGR)
        lightness = lab_img[:, :, 0]
        green_magenta = lab_img[:, :, 1]
        blue_yellow = lab_img[:, :, 2]

        # Make a mask for each channel (Use '*' to avoid linter error about unbalanced value unpacking)
        *out_masks, = _call_inrange(gray_imgs_list=[lightness, green_magenta, blue_yellow],
                                    lower_thresh=lower_thresh, upper_thresh=upper_thresh)

        # Apply the masks to the image
        result = cv2.bitwise_and(img, img, mask=out_masks[0])
        result = cv2.bitwise_and(result, result, mask=out_masks[1])
        masked_img = cv2.bitwise_and(result, result, mask=out_masks[2])

        # Combine masks
        mask = cv2.bitwise_and(out_masks[0], out_masks[1])
        mask = cv2.bitwise_and(mask, out_masks[2])

    elif channel.upper() == 'CMYK':

        # Check threshold inputs
        if not (len(lower_thresh) == 4 and len(upper_thresh) == 4):
            fatal_error("If using the CYMK colorspace, 4 thresholds are needed for both lower_thresh and " +
                        "upper_thresh. If thresholding isn't needed for a channel, set lower_thresh=0 and " +
                        "upper_thresh=255")

        # Convert the RGB image to LAB colorspace
        c = _rgb2cmyk(rgb_img=img, channel="c")
        m = _rgb2cmyk(rgb_img=img, channel="m")
        y = _rgb2cmyk(rgb_img=img, channel="y")
        k = _rgb2cmyk(rgb_img=img, channel="k")

        # Make a mask for each channel
        *out_masks, = _call_inrange(gray_imgs_list=[c, m, y, k], lower_thresh=lower_thresh, upper_thresh=upper_thresh)
        # Use '*' to avoid linter error about unbalanced value unpacking

        # Apply the masks to the image
        result = cv2.bitwise_and(img, img, mask=out_masks[0])
        result = cv2.bitwise_and(result, result, mask=out_masks[1])
        result = cv2.bitwise_and(result, result, mask=out_masks[2])
        masked_img = cv2.bitwise_and(result, result, mask=out_masks[3])

        # Combine masks
        mask = cv2.bitwise_and(out_masks[0], out_masks[1])
        mask2 = cv2.bitwise_and(out_masks[2], out_masks[3])
        mask = cv2.bitwise_and(mask, mask2)

    elif channel.upper() in ('GRAY', 'GREY'):

        # Check threshold input
        _check_threshold_inputs(1, lower_thresh, upper_thresh)

        if len(np.shape(img)) == 3:
            # Convert RGB image to grayscale colorspace
            gray_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        else:
            gray_img = img

        # Make a mask
        mask = cv2.inRange(gray_img, lower_thresh[0], upper_thresh[0])

        # Apply the masks to the image
        masked_img = cv2.bitwise_and(img, img, mask=mask)

    else:
        fatal_error(str(channel) + " is not a valid colorspace. Channel must be either 'RGB', 'HSV', 'CMYK', or 'gray'.")

    # Auto-increment the device counter

    # Print or plot the binary image if debug is on
    _debug(visual=masked_img, filename=os.path.join(params.debug_outdir,
                                                    str(params.device) + channel + 'custom_thresh.png'))
    _debug(visual=mask, filename=os.path.join(params.debug_outdir,
                                              str(params.device) + channel + 'custom_thresh_mask.png'))
    return mask, masked_img


# Internal method for calling the OpenCV inRange function to reduce code duplication
def _call_inrange(gray_imgs_list, lower_thresh, upper_thresh):
    """Calls the OpenCV threshold function to reduce code duplication

    Parameters
    ----------
    gray_imgs_list : list
        List containing grayscale arrays
    lower_thresh : list
        List of lower threshold values (0-255)
    upper_thresh : list
        List of upper threshold values (0-255)

    Returns
    -------
    list
        Thresholded, binary images
    """
    out_masks = []
    # Apply inRange to each array in the list
    for i, array in enumerate(gray_imgs_list):
        mask = cv2.inRange(array, lower_thresh[i], upper_thresh[i])
        out_masks.append(mask)
    return out_masks


# Internal method for checking threshold inputs to reduce code duplication
def _check_threshold_inputs(required_length, lower_thresh, upper_thresh):
    """Checks threshold inputs for correct number of inputs

    Parameters
    ----------
    required_length : int
        Number of required inputs
    lower_thresh : list
        List of lower threshold values (0-255)
    upper_thresh : list
        List of upper threshold values (0-255)

    Raises
    -------
    RuntimeError
        If the number of threshold inputs is incorrect
    """
    if not (len(lower_thresh) == required_length and len(upper_thresh) == required_length):
        fatal_error(f"{required_length} threshold inputs are needed for both lower_thresh and " +
                    "upper_thresh for the given channel. If thresholding isn't needed for a channel, set lower_thresh=0 and " +
                    "upper_thresh=255")


# Internal method for calling the OpenCV threshold function to reduce code duplication
def _call_threshold(gray_img, threshold, threshold_method, method_name):
    """Calls the OpenCV threshold function to reduce code duplication

    Parameters
    ----------
    gray_img : numpy.ndarray
        Grayscale image data
    threshold : int
        Threshold value
    threshold_method : int
        OpenCV thresholding method
    method_name : str
        Name of the method used for debugging purposes

    Returns
    -------
    numpy.ndarray
        Thresholded, binary image
    """
    # Threshold the image
    _, bin_img = cv2.threshold(gray_img, threshold, 255, threshold_method)

    if bin_img.dtype != 'uint16':
        bin_img = np.uint8(bin_img)

    # Print or plot the binary image if debug is on
    _debug(visual=bin_img, filename=os.path.join(params.debug_outdir,
                                                 str(params.device) + method_name + str(threshold) + '.png'))

    return bin_img


# Internal method for calling the OpenCV adaptiveThreshold function to reduce code duplication
def _call_adaptive_threshold(gray_img, ksize, offset, adaptive_method, threshold_method, method_name):
    """Calls the OpenCV adaptiveThreshold function to reduce code duplication

    Parameters
    ----------
    gray_img : numpy.ndarray
        Grayscale image data
    ksize : int
        Size of the block of pixels used to compute the local average
    offset : float
        Value substracted from the local average to compute the local threshold
    adaptive_method : int
        Adaptive thresholding algorithm to use
    threshold_method : int
        Thresholding method to use
    method_name : str
        Name of the method used for debugging purposes

    Returns
    -------
    numpy.ndarray
        Thresholded, binary image
    """
    if ksize < 3:
        fatal_error("ksize must be >= 3")

    # Force ksize to be odd number
    ksize = int(ksize)
    if (ksize % 2) != 1:
        ksize = ksize + 1

    # Threshold the image
    bin_img = cv2.adaptiveThreshold(gray_img, 255, adaptive_method, threshold_method, ksize, offset)

    # Print or plot the binary image if debug is on
    _debug(visual=bin_img, filename=os.path.join(params.debug_outdir, str(params.device) + method_name + '.png'))

    return bin_img


# Internal method for detecting peaks for the triangle autothreshold method
def _detect_peaks(x, mph=None, mpd=1, threshold=0, edge='rising', valley=False, show=False, ax=None):
    """Marcos Duarte, https://github.com/demotu/BMC; version 1.0.4; license MIT

    Detect peaks in data based on their amplitude and other features.

    Parameters
    ----------
    x : 1D array_like
        data.
    mph : {None, number}, optional (default = None)
        detect peaks that are greater than minimum peak height.
    mpd : positive integer, optional (default = 1)
        detect peaks that are at least separated by minimum peak distance (in
        number of data).
    threshold : positive number, optional (default = 0)
        detect peaks (valleys) that are greater (smaller) than `threshold`
        in relation to their immediate neighbors.
    edge : {None, 'rising', 'falling', 'both'}, optional (default = 'rising')
        for a flat peak, keep only the rising edge ('rising'), only the
        falling edge ('falling'), both edges ('both'), or don't detect a
        flat peak (None).
    valley : bool, optional (default = False)
        if True (1), detect valleys (local minima) instead of peaks.
    show : bool, optional (default = False)
        if True (1), plot data in matplotlib figure.
    ax : a matplotlib.axes.Axes instance, optional (default = None).

    Returns
    -------
    ind : 1D array_like
        indices of the peaks in `x`.

    Notes
    -----
    The detection of valleys instead of peaks is performed internally by simply
    negating the data: `ind_valleys = detect_peaks(-x)`

    The function can handle NaN's

    See this IPython Notebook [1]_.

    References
    ----------
    .. [1] http://nbviewer.ipython.org/github/demotu/BMC/blob/master/notebooks/DetectPeaks.ipynb

    Examples
    --------
    from detect_peaks import detect_peaks
    x = np.random.randn(100)
    x[60:81] = np.nan
    # detect all peaks and plot data
    ind = detect_peaks(x, show=True)
    print(ind)

    x = np.sin(2*np.pi*5*np.linspace(0, 1, 200)) + np.random.randn(200)/5
    # set minimum peak height = 0 and minimum peak distance = 20
    detect_peaks(x, mph=0, mpd=20, show=True)

    x = [0, 1, 0, 2, 0, 3, 0, 2, 0, 1, 0]
    # set minimum peak distance = 2
    detect_peaks(x, mpd=2, show=True)

    x = np.sin(2*np.pi*5*np.linspace(0, 1, 200)) + np.random.randn(200)/5
    # detection of valleys instead of peaks
    detect_peaks(x, mph=0, mpd=20, valley=True, show=True)

    x = [0, 1, 1, 0, 1, 1, 0]
    # detect both edges
    detect_peaks(x, edge='both', show=True)

    x = [-2, 1, -2, 2, 1, 1, 3, 0]
    # set threshold = 2
    detect_peaks(x, threshold = 2, show=True)
    """
    x = np.atleast_1d(x).astype('float64')

    # It is always the case that x.size=256 since 256 hardcoded in line 186 ->
    # cv2.calcHist([gray_img], [0], None, [256], [0, 255])
    # if x.size < 3:
    #     return np.array([], dtype=int)

    # # Where this function is used it is hardcoded to use the default valley=False so this will never be used
    # if valley:
    #     x = -x
    # find indices of all peaks
    dx = x[1:] - x[:-1]
    # handle NaN's
    # indnan = np.where(np.isnan(x))[0]

    # x will never contain NaN since calcHist will never return NaN
    # if indnan.size:
    #     x[indnan] = np.inf
    #     dx[np.where(np.isnan(dx))[0]] = np.inf
    ine, ire, ife = np.array([[], [], []], dtype=int)

    if edge.lower() in ['rising', 'both']:
        ire = np.where((np.hstack((dx, 0)) <= 0) & (np.hstack((0, dx)) > 0))[0]
    ind = np.unique(np.hstack((ine, ire, ife)))
    # x will never contain NaN since calcHist will never return NaN
    # if ind.size and indnan.size:
    #     # NaN's and values close to NaN's cannot be peaks
    #     ind = ind[np.in1d(ind, np.unique(np.hstack((indnan, indnan - 1, indnan + 1))), invert=True)]
    # first and last values of x cannot be peaks
    # if ind.size and ind[0] == 0:
    #     ind = ind[1:]
    # if ind.size and ind[-1] == x.size - 1:
    #     ind = ind[:-1]
    # We think the above code will never be reached given some of the hardcoded properties used

    # # Where this function is used has hardcoded mph=None so this will never be used
    # # remove peaks < minimum peak height
    # if ind.size and mph is not None:
    #     ind = ind[x[ind] >= mph]
    # remove peaks - neighbors < threshold

    if show:
        # x will never contain NaN since calcHist will never return NaN
        # if indnan.size:
        #     x[indnan] = np.nan
        # # Where this function is used it is hardcoded to use the default valley=False so this will never be used
        # if valley:
        #     x = -x
        _plot(x, mph, mpd, threshold, edge, valley, ax, ind)

    return ind


# Internal plotting function for the triangle autothreshold method
def _plot(x, mph, mpd, threshold, edge, valley, ax, ind):
    """Plot results of the detect_peaks function, see its help."""
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(8, 4))

    ax.plot(x, 'b', lw=1)
    if ind.size:
        label = 'valley' if valley else 'peak'
        label = label + 's' if ind.size > 1 else label
        ax.plot(ind, x[ind], '+', mfc=None, mec='r', mew=2, ms=8,
                label=f'{ind.size} {label}')
        ax.legend(loc='best', framealpha=.5, numpoints=1)
    ax.set_xlim(-.02 * x.size, x.size * 1.02 - 1)
    ymin, ymax = x[np.isfinite(x)].min(), x[np.isfinite(x)].max()
    yrange = ymax - ymin if ymax > ymin else 1
    ax.set_ylim(ymin - 0.1 * yrange, ymax + 0.1 * yrange)
    ax.set_xlabel('Data #', fontsize=14)
    ax.set_ylabel('Amplitude', fontsize=14)
    mode = 'Valley detection' if valley else 'Peak detection'
    ax.set_title(f"{mode} ({mph=}, {mpd=}, {threshold=}, {edge=})")
    plt.show()


def saturation(rgb_img, threshold=255, channel="any"):
    """Return a mask filtering out saturated pixels.

    Inputs:
    rgb_img    = RGB image
    threshold  = value for threshold, above which is considered saturated
    channel    = how many channels must be saturated for the pixel to be masked out ("any", "all")

    Returns:
    masked_img = A binary image with the saturated regions blacked out.

    :param rgb_img: np.ndarray
    :param threshold: int
    :param channel: str
    :return masked_img: np.ndarray
    """
    # Mask red, green, and blue saturation separately
    b, g, r = cv2.split(rgb_img)
    b_saturated = cv2.inRange(b, threshold, 255)
    g_saturated = cv2.inRange(g, threshold, 255)
    r_saturated = cv2.inRange(r, threshold, 255)

    # Combine channel masks
    if channel.lower() == "any":
        # Consider a pixel saturated if any channel is saturated
        saturated = cv2.bitwise_or(b_saturated, g_saturated)
        saturated = cv2.bitwise_or(saturated, r_saturated)
    elif channel.lower() == "all":
        # Consider a pixel saturated only if all channels are saturated
        saturated = cv2.bitwise_and(b_saturated, g_saturated)
        saturated = cv2.bitwise_and(saturated, r_saturated)
    else:
        fatal_error(str(channel) + " is not a valid option. Channel must be either 'any', or 'all'.")

    # Invert "saturated" before returning, so saturated = black
    bin_img = cv2.bitwise_not(saturated)

    _debug(visual=bin_img, filename=os.path.join(params.debug_outdir, str(params.device), '_saturation_threshold.png'))
    return bin_img


def mask_bad(float_img, bad_type='native'):
    """Create a mask with desired "bad" pixels of the input floaat image marked.
    Inputs:
    float_img = image represented by an nd-array (data type: float). Most probably, it is the result of some
                calculation based on the original image. So the datatype is float, and it is possible to have some
                "bad" values, i.e. nan and/or inf
    bad_type = definition of "bad" type, can be 'nan', 'inf' or 'native'
    Returns:
    mask = A mask indicating the locations of "bad" pixels

    :param float_img: numpy.ndarray
    :param bad_type: str
    :return mask: numpy.ndarray
    """
    size_img = np.shape(float_img)
    if len(size_img) != 2:
        fatal_error('Input image is not a single channel image!')

    mask = np.zeros(size_img, dtype='uint8')
    idx_nan, idy_nan = np.where(np.isnan(float_img) == 1)
    idx_inf, idy_inf = np.where(np.isinf(float_img) == 1)

    # neither nan nor inf exists in the image, print out a message and the mask would just be all zero
    if len(idx_nan) == 0 and len(idx_inf) == 0:
        print('Neither nan nor inf appears in the current image.')
    # at least one of the "bad" exists
    # desired bad to mark is "native"
    elif bad_type.lower() == 'native':
        mask[idx_nan, idy_nan] = 255
        mask[idx_inf, idy_inf] = 255
    elif bad_type.lower() == 'nan' and len(idx_nan) >= 1:
        mask[idx_nan, idy_nan] = 255
    elif bad_type.lower() == 'inf' and len(idx_inf) >= 1:
        mask[idx_inf, idy_inf] = 255
    # "bad" exists but not the user desired bad type, return the all-zero mask
    else:
        print(f'{format(bad_type.lower())} does not appear in the current image.')

    _debug(visual=mask, filename=os.path.join(params.debug_outdir, str(params.device) + "_bad_mask.png"))

    return mask


# functions to get a given channel with parameters compatible
# with rgb2gray_lab, rgb2gray_hsv, and rgb2gray_cmyk to use in the dict
def _get_R(rgb_img, _):
    """Get the red channel from a RGB image"""
    return rgb_img[:, :, 2]


def _get_G(rgb_img, _):
    """Get the green channel from a RGB image"""
    return rgb_img[:, :, 1]


def _get_B(rgb_img, _):
    """Get the blue channel from a RGB image"""
    return rgb_img[:, :, 0]


def _get_gray(rgb_img, _):
    """Get the gray scale transformation of a RGB image"""
    return _rgb2gray(rgb_img=rgb_img)


def _get_index(rgb_img, _):
    """Get a vector with linear indices of the pixels in an image"""
    h, w, _ = rgb_img.shape
    return np.arange(h*w).reshape(h, w)


def _not_valid(*args):
    """Error for a non valid channel"""
    return fatal_error("channel not valid, use R, G, B, l, a, b, h, s, v, c, m, y, k, gray, or index")


def dual_channels(rgb_img, x_channel, y_channel, points, above=True):
    """
    Create a binary image from an RGB image based on the pixels values in two channels.
    The x and y channels define a 2D plane and the two input points define a straight line.
    Pixels in the plane above and below the straight line are assigned two different values.

    Inputs:
    rgb_img   = RGB image
    x_channel = Channel to use for the horizontal coordinate.
                Options:  'R', 'G', 'B', 'l', 'a', 'b', 'h', 's', 'v', 'c', 'm', 'y', 'k', 'gray', and 'index'
    y_channel = Channel to use for the vertical coordinate.
                Options:  'R', 'G', 'B', 'l', 'a', 'b', 'h', 's', 'v', 'c', 'm', 'y', 'k', 'gray', and 'index'
    points    = List containing two points as tuples defining the segmenting straight line
    above     = Whether the pixels above the line are given the value of 0 or max_value

    Returns:
    bin_img = Thresholded, binary image

    :param rgb_img: numpy.ndarray
    :param x_channel: str
    :param y_channel: str
    :param points: list of two tuples
    :param above: bool
    :return bin_img: numpy.ndarray
    """
    # dictionary returns the function that gets the required image channel
    channel_dict = {
        'R': _get_R,
        'G': _get_G,
        'B': _get_B,
        'l': _rgb2lab,
        'a': _rgb2lab,
        'b': _rgb2lab,
        'gray': _get_gray,
        'h': _rgb2hsv,
        's': _rgb2hsv,
        'v': _rgb2hsv,
        'index': _get_index,
        'c': _rgb2cmyk,
        'm': _rgb2cmyk,
        'y': _rgb2cmyk,
        'k': _rgb2cmyk
    }

    debug = params.debug
    params.debug = None
    # get channels
    img_x_ch = channel_dict.get(x_channel, _not_valid)(rgb_img, x_channel)
    img_x_ch = img_x_ch.astype(np.float64)
    img_y_ch = channel_dict.get(y_channel, _not_valid)(rgb_img, y_channel)
    img_y_ch = img_y_ch.astype(np.float64)
    params.debug = debug

    if len(points) < 2:
        fatal_error('Two points are required')

    if len(points) > 2:
        # Print warning statement
        warn("only the first two points are used in this function")

    x0, y0 = points[0]
    x1, y1 = points[1]

    m = (y1-y0) / (x1-x0+1e-10)  # avoid division by 0
    b = y0 - m*x0

    y_line = m*img_x_ch + b

    max_value = 255
    if above:
        bin_img = max_value*(img_y_ch > y_line)
    else:
        bin_img = max_value*(img_y_ch < y_line)

    bin_img = bin_img.astype(np.uint8)

    _debug(visual=bin_img, filename=os.path.join(params.debug_outdir,
                                                 str(params.device) + '_' + x_channel + y_channel + '_2D_threshold_mask.png'))

    return bin_img



"""Script to generate an image with metrics displayed on image."""
import numpy as np
import math
import numbers
from plantcv.plantcv import params
from plantcv.plantcv import outputs


def landmark_reference_pt_dist(points_r, centroid_r, bline_r, label=None):
    """landmark_reference_pt_dist

    For each point in contour, get a point before (pre) and after (post) the point of interest.
    The win argument specifies the pre and post point distances.

    Inputs:
    points_r   = a set of rescaled points (basically the output of the acute_vertex fxn after the scale_features fxn)
    centroid_r = a tuple that contains the rescaled centroid coordinates
    bline_r    = a tuple that contains the rescaled boundary line - centroid coordinates
    label      = Optional label parameter, modifies the variable name of
                 observations recorded (default = pcv.params.sample_label).

    :param points_r: ndarray
    :param centroid_r: tuple
    :param bline_r: tuple
    :param label: str
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    params.device += 1
    vert_dist_c = []
    hori_dist_c = []
    euc_dist_c = []
    angles_c = []
    cx, cy = centroid_r
    # Check to see if points are numerical or NA
    if not isinstance(cy, numbers.Number):
        return ('NA', 'NA'), ('NA', 'NA'), ('NA', 'NA'), ('NA', 'NA'), ('NA', 'NA'), ('NA', 'NA'), \
               ('NA', 'NA'), ('NA', 'NA')
    # Do this for centroid
    for pt in points_r:
        # Get coordinates from point
        x, y = pt
        # Get vertical distance and append to list
        v = y - cy
        vert_dist_c.append(v)

        # Get horizontal distance and append to list
        h = abs(x - cx)
        hori_dist_c.append(h)
        e = np.sqrt((cx - x) * (cx - x) + (cy - y) * (cy - y))
        # print "Here is the centroid euclidian distance: " + str(h)
        euc_dist_c.append(e)

        a = (h * h + e * e - v * v) / (2 * h * e)

        ang = abs(math.degrees(math.acos(a)))
        if v < 0:
            ang = ang * -1
        # print "Here is the centroid angle: " + str(ang)
        angles_c.append(ang)

    vert_ave_c = np.mean(vert_dist_c)
    hori_ave_c = np.mean(hori_dist_c)
    euc_ave_c = np.mean(euc_dist_c)
    ang_ave_c = np.mean(angles_c)

    vert_dist_b = []
    hori_dist_b = []
    euc_dist_b = []
    angles_b = []
    bx, by = bline_r
    # Do this for baseline
    for pt in points_r:
        # Get coordinates from point
        x, y = pt
        # Get vertical distance and append to list
        v = y - by
        # print "Here is the baseline vertical distance: " + str(v)
        vert_dist_b.append(v)
        # Get horizontal distance and append to list
        h = abs(x - bx)
        hori_dist_b.append(h)
        e = np.sqrt((bx - x) * (bx - x) + (by - y) * (by - y))
        euc_dist_b.append(e)

        a = (h * h + e * e - v * v) / (2 * h * e)

        ang = abs(math.degrees(math.acos(a)))
        if v < 0:
            ang = ang * -1
        angles_b.append(ang)

    vert_ave_b = np.mean(vert_dist_b)
    hori_ave_b = np.mean(hori_dist_b)
    euc_ave_b = np.mean(euc_dist_b)
    ang_ave_b = np.mean(angles_b)

    outputs.add_observation(sample=label, variable='vert_ave_c', trait='average vertical distance from centroid',
                            method='plantcv.plantcv.homology.landmark_reference_pt_dist', scale='pixels', datatype=float,
                            value=vert_ave_c, label='pixels')
    outputs.add_observation(sample=label, variable='hori_ave_c', trait='average horizontal distance from centeroid',
                            method='plantcv.plantcv.homology.landmark_reference_pt_dist', scale='pixels', datatype=float,
                            value=hori_ave_c, label='pixels')
    outputs.add_observation(sample=label, variable='euc_ave_c', trait='average euclidean distance from centroid',
                            method='plantcv.plantcv.homology.landmark_reference_pt_dist', scale='pixels', datatype=float,
                            value=euc_ave_c, label='pixels')
    outputs.add_observation(sample=label, variable='ang_ave_c',
                            trait='average angle between landmark point and centroid',
                            method='plantcv.plantcv.homology.landmark_reference_pt_dist', scale='degrees', datatype=float,
                            value=ang_ave_c, label='degrees')
    outputs.add_observation(sample=label, variable='vert_ave_b', trait='average vertical distance from baseline',
                            method='plantcv.plantcv.homology.landmark_reference_pt_dist', scale='pixels', datatype=float,
                            value=vert_ave_b, label='pixels')
    outputs.add_observation(sample=label, variable='hori_ave_b', trait='average horizontal distance from baseline',
                            method='plantcv.plantcv.homology.landmark_reference_pt_dist', scale='pixels', datatype=float,
                            value=hori_ave_b, label='pixels')
    outputs.add_observation(sample=label, variable='euc_ave_b', trait='average euclidean distance from baseline',
                            method='plantcv.plantcv.homology.landmark_reference_pt_dist', scale='pixels', datatype=float,
                            value=euc_ave_b, label='pixels')
    outputs.add_observation(sample=label, variable='ang_ave_b',
                            trait='average angle between landmark point and baseline',
                            method='plantcv.plantcv.homology.landmark_reference_pt_dist', scale='degrees', datatype=float,
                            value=ang_ave_b, label='degrees')
    return None



"""Visualize histograms from image data."""
import os
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import fatal_error
from plantcv.plantcv._debug import _debug
import pandas as pd
import altair as alt


def _hist_gray(gray_img, bins, lower_bound, upper_bound, mask=None):
    """Prepare the ready to plot histogram data.

    Inputs:
    gray_img       = grayscale image to analyze
    bins           = divide the data into n evenly spaced bins
    lower_bound    = the lower bound of the bins (x-axis min value)
    upper_bound    = the upper bound of the bins (x-axis max value)
    mask           = binary mask, calculate histogram from masked area only (default=None)

    Returns:
    bin_labels     = an array of histogram bin labels
    hist_percent   = an array of histogram represented by percent values
    hist_gray_data = an array of histogram (original values)

    :param gray_img: numpy.ndarray
    :param bins: int
    :param lower_bound: int
    :param upper_bound: int
    :param mask: numpy.ndarray
    :return bin_labels: numpy.ndarray
    :return hist_percent: numpy.ndarray
    :return hist_gray_data: numpy.ndarray
    """
    params.device += 1
    # Create a dummy mask if none was supplied
    if mask is None:
        mask = np.ones(gray_img.shape, dtype=np.uint8)

    # Apply mask
    pixels = len(np.where(mask > 0)[0])
    masked = gray_img[np.where(mask > 0)]

    # Store histogram data
    hist_gray_data, hist_bins = np.histogram(masked, bins, (lower_bound, upper_bound))

    # make hist percentage for plotting
    hist_percent = (hist_gray_data / float(pixels)) * 100
    # use middle value of every bin as bin label
    bin_labels = np.array([np.average([hist_bins[i], hist_bins[i+1]]) for i in range(0, len(hist_bins) - 1)])

    return bin_labels, hist_percent, hist_gray_data
    # hist_data = pd.DataFrame({'pixel intensity': bin_labels, 'proportion of pixels (%)': hist_percent})
    # return hist_data


def _get_histogram_bounds(img, mask=None, upper_bound=None, lower_bound=None):
    """
    Get boundaries for where to count histogram pixels.

    Parameters
    ----------
    img : numpy.ndarray
        An RGB or grayscale image.
    mask : numpy.ndarray, optional
        A binary mask image.
    upper_bound : int, optional
        Upper bound for bins.
    lower_bound : int, optional
        Lower bound for bins.

    Returns
    -------
    upper_bound : int
        New upper boundary for bins.
    lower_bound : int
        New lower boundary for bins.
    """
    img_min, img_max = np.nanmin(img), np.nanmax(img)
    if mask is not None:
        masked = img[np.where(mask > 0)]
        img_min, img_max = np.nanmin(masked), np.nanmax(masked)
    # for lower / upper bound, if given, use the given value, otherwise, use the min / max of the image
    if lower_bound is None:
        lower_bound = img_min
    if upper_bound is None:
        upper_bound = img_max
    return upper_bound, lower_bound


def histogram(img, mask=None, bins=100, lower_bound=None, upper_bound=None, title=None, hist_data=False):
    """Plot histograms of each input image channel.

    Inputs:
    img            = an RGB or grayscale image to analyze
    mask           = binary mask, calculate histogram from masked area only (default=None)
    bins           = divide the data into n evenly spaced bins (default=100)
    lower_bound    = the lower bound of the bins (x-axis min value) (default=None)
    upper_bound    = the upper bound of the bins (x-axis max value) (default=None)
    title          = a custom title for the plot (default=None)
    hist_data      = return the frequency distribution data if True (default=False)

    Returns:
    chart          = histogram figure
    hist_df        = dataframe with histogram data, with columns "pixel intensity" and "proportion of pixels (%)"

    :param img: numpy.ndarray
    :param mask: numpy.ndarray
    :param bins: int
    :param lower_bound: int
    :param upper_bound: int
    :param title: str
    :param hist_data: bool
    :return chart: altair.vegalite.v5.api.Chart
    :return hist_df: pandas.core.frame.DataFrame
    """
    if not isinstance(img, np.ndarray):
        fatal_error("Only image of type numpy.ndarray is supported input!")
    if len(img.shape) < 2:
        fatal_error("Input image should be at least a 2d array!")

    upper_bound, lower_bound = _get_histogram_bounds(img, mask, upper_bound, lower_bound)

    if len(img.shape) == 2:
        bin_labels, hist_percent, hist_ = _hist_gray(img, bins=bins, lower_bound=lower_bound, upper_bound=upper_bound,
                                                     mask=mask)
        hist_df = pd.DataFrame(
            {'pixel intensity': bin_labels, 'proportion of pixels (%)': hist_percent, 'hist_count': hist_,
             'color channel': ['0' for _ in range(len(hist_percent))]})
    # otherwise must be greater than 2 image shape
    else:
        b_names = [str(i) for i in range(img.shape[2])]
        if img.shape[2] == 3:
            b_names = ['blue', 'green', 'red']
        # Assumption: RGB image
        # Initialize dataframe column arrays
        px_int = np.array([])
        prop = np.array([])
        hist_count = np.array([])
        channel = []
        for (b, b_name) in enumerate(b_names):
            bin_labels, hist_percent, hist_ = _hist_gray(img[:, :, b], bins=bins, lower_bound=lower_bound,
                                                         upper_bound=upper_bound, mask=mask)
            # Append histogram data for each channel
            px_int = np.append(px_int, bin_labels)
            prop = np.append(prop, hist_percent)
            hist_count = np.append(hist_count, hist_)
            channel = channel + [b_name for _ in range(len(hist_percent))]
        # Create dataframe
        hist_df = pd.DataFrame(
            {'pixel intensity': px_int, 'proportion of pixels (%)': prop, 'hist_count': hist_count,
             'color channel': channel})

    # Create an altair chart
    chart = alt.Chart(hist_df).mark_line(point=True).encode(
        x="pixel intensity",
        y="proportion of pixels (%)",
        color="color channel",
        tooltip=['pixel intensity', 'proportion of pixels (%)']
        )

    if title is not None:
        chart = chart.properties(title=title)

    if len(img.shape) > 2 and img.shape[2] == 3:
        # Add a blue, green, red color scale if the image is RGB
        chart = chart.configure_range(category=['blue', 'green', 'red'])

    # Plot or print the histogram
    _debug(visual=chart, filename=os.path.join(params.debug_outdir, str(params.device) + '_hist.png'))

    if hist_data is True:
        return chart.interactive(), hist_df
    return chart.interactive()



# Plot segment ID numbers

import os
import cv2
from plantcv.plantcv import color_palette
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug


def segment_id(skel_img, objects, mask=None):
    """Plot segment IDs.

    Inputs:
    skel_img      = Skeletonized image
    objects       = List of contours
    mask          = (Optional) binary mask for debugging. If provided, debug image will be overlaid on the mask.

    Returns:
    segmented_img = Segmented image
    labeled_img   = Labeled image

    :param skel_img: numpy.ndarray
    :param objects: list
    :param mask: numpy.ndarray
    :return segmented_img: numpy.ndarray
    :return labeled_img: numpy.ndarray
    """
    label_coord_x = []
    label_coord_y = []

    if mask is None:
        segmented_img = skel_img.copy()
    else:
        segmented_img = mask.copy()

    segmented_img = cv2.cvtColor(segmented_img, cv2.COLOR_GRAY2RGB)

    # Create a color scale, use a previously stored scale if available
    rand_color = color_palette(num=len(objects), saved=True)

    # Plot all segment contours
    for i, cnt in enumerate(objects):
        cv2.drawContours(segmented_img, cnt, -1, rand_color[i], params.line_thickness, lineType=8)
        # Store coordinates for labels
        label_coord_x.append(objects[i][0][0][0])
        label_coord_y.append(objects[i][0][0][1])

    labeled_img = segmented_img.copy()

    for i, cnt in enumerate(objects):
        # Label slope lines
        w = label_coord_x[i]
        h = label_coord_y[i]
        text = f"{i}"
        cv2.putText(img=labeled_img, text=text, org=(w, h), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=params.text_size, color=rand_color[i], thickness=params.text_thickness)

    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_segmented_ids.png"))

    return segmented_img, labeled_img



# Visualize an RGB image in all potential colorspaces as one glance

import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv.transform import resize_factor
from plantcv.plantcv import fatal_error
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rgb2lab, _rgb2hsv, _rgb2cmyk


def colorspaces(rgb_img, original_img=True):
    """Visualize an RGB image in all potential colorspaces.

    Inputs:
    rgb_img      = RGB image data
    original_img = Whether or not to include the original image the the debugging plot

    Returns:
    plotting_img = Plotting image containing the original image and L,A,B,H,S,V,C,M,Y, and K colorspaces

    :param rgb_img: numpy.ndarray
    :param original_img: bool
    :return labeled_img: numpy.ndarray
    """
    if not len(np.shape(rgb_img)) == 3:
        fatal_error("Input image is not RGB!")

    # Store and disable debug mode
    debug = params.debug
    params.debug = None

    # Initialize grayscale images list, rgb images list, plotting coordinates
    colorspace_names = ["H", "S", "V", "L", "A", "B", "C", "M", "Y", "K"]
    all_colorspaces = []
    labeled_imgs = []
    y = int(np.shape(rgb_img)[0] / 2)
    x = int(np.shape(rgb_img)[1] / 2)

    # Loop through and create grayscale imgs from each colorspace
    for i in range(0, 3):
        channel = colorspace_names[i]
        all_colorspaces.append(_rgb2hsv(rgb_img=rgb_img, channel=channel))
    for i in range(3, 6):
        channel = colorspace_names[i]
        all_colorspaces.append(_rgb2lab(rgb_img=rgb_img, channel=channel))
    for i in range(6, 10):
        channel = colorspace_names[i]
        all_colorspaces.append(_rgb2cmyk(rgb_img=rgb_img, channel=channel))

    # Plot labels of each colorspace on the corresponding img
    for i, colorspace in enumerate(all_colorspaces):
        converted_img = cv2.cvtColor(colorspace, cv2.COLOR_GRAY2RGB)
        labeled = cv2.putText(img=converted_img, text=colorspace_names[i], org=(x, y),
                              fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                              fontScale=params.text_size, color=(255, 0, 255), thickness=params.text_thickness)
        labeled_imgs.append(labeled)

    # Resize and stack CMYK gray imgs to be able to vertically stack
    cmky_labeled_img = np.hstack([resize_factor(labeled_imgs[6], factors=(3/4, 3/4)),
                                 resize_factor(labeled_imgs[7], factors=(3/4, 3/4)),
                                 resize_factor(labeled_imgs[8], factors=(3/4, 3/4)),
                                 resize_factor(labeled_imgs[9], factors=(3/4, 3/4))])

    # Compile images together, and add padding to resize so that image widths are all equal
    sub_imgs = [np.hstack([labeled_imgs[0], labeled_imgs[1], labeled_imgs[2]]),
                np.hstack([labeled_imgs[3], labeled_imgs[4], labeled_imgs[5]]),
                cmky_labeled_img]
    img_sizes = []
    for img in sub_imgs:
        width = img.shape[1]
        img_sizes.append(width)
    max_width = max(img_sizes)
    new = [cv2.copyMakeBorder(arr, 0, 0, (max_width - arr.shape[1]), 0,
                              cv2.BORDER_CONSTANT, value=(0, 0, 0)) for arr in sub_imgs]
    plotting_img = np.vstack(new)

    # If original_img is True then also plot the original image with the rest of them
    if original_img:
        plotting_img = np.hstack([resize_factor(img=rgb_img, factors=(3, 11/4)), plotting_img])
    plotting_img = resize_factor(plotting_img, factors=(0.5, 0.5))

    # Reset debug mode
    params.debug = debug

    _debug(visual=plotting_img, filename=os.path.join(params.debug_outdir, f"{params.device}_vis_colorspaces.png"))

    return plotting_img



# Create labels base on clean mask and optionally, multiple ROIs
import cv2
import os
import numpy as np
from skimage.measure import label
from skimage.color import label2rgb
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _roi_filter, _cv2_findcontours


def create_labels(mask, rois=None, roi_type="partial"):
    """Create a labeled mask where connected regions of non-zero
    pixels are assigned a label value based on the provided
    region of interest (ROI).

    Inputs:
    mask            = mask image
    rois            = list of multiple ROIs (from roi.multi or roi.auto_grid)
    roi_type        = 'cutto', 'partial' (for partially inside, default),
                    'largest' (keep only the largest contour), or 'auto'
                    (use the mask alone withtout roi filtering)

    Returns:
    mask            = Labeled mask
    num_labels      = Number of labeled objects

    :param mask: numpy.ndarray
    :param rois: plantcv.plantcv.classes.Objects
    :return labeled_mask: numpy.ndarray
    :return num_labels: int
    """
    # Store debug mode
    debug = params.debug
    params.debug = None

    # Use contours for labeling
    if rois is None:
        # label will work with any number of objects in the mask
        labeled_mask, num_labels = label(mask, background=0,
                                         return_num=True, connectivity=2)

    # Use the rois for labeling
    else:
        contours, hierarchy = _cv2_findcontours(mask)
        labeled_mask = np.zeros(mask.shape[:2], dtype=np.int32)
        num_labels = len(rois.contours)
        for i, roi in enumerate(rois):
            kept_cnt, _, mask = _roi_filter(img=mask, roi=roi,
                                            obj=contours,
                                            hierarchy=hierarchy,
                                            roi_type=roi_type)

            # Pixel intensity of (i+1) such that the first object has value
            cv2.drawContours(labeled_mask, kept_cnt, -1, (i+1), -1)

    # Restore debug parameter
    params.debug = debug
    colorful = label2rgb(labeled_mask)
    colorful2 = (255*colorful).astype(np.uint8)

    _debug(colorful2, filename=os.path.join(params.debug_outdir,
                                            str(params.device) +
                                            '_label_colored_mask.png'))

    return labeled_mask, num_labels



# Image subtraction

from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _image_subtract
from plantcv.plantcv import params
import os


def image_subtract(gray_img1, gray_img2):
    """
    This is a function used to subtract values of one gray-scale image array from another gray-scale image array. The
    resulting gray-scale image array has a minimum element value of zero. That is all negative values resulting from the
    subtraction are forced to zero.

    Inputs:
    gray_img1   = Grayscale image data from which gray_img2 will be subtracted
    gray_img2   = Grayscale image data which will be subtracted from gray_img1

    Returns:
    new_img = subtracted image

    :param gray_img1: numpy.ndarray
    :param gray_img2: numpy.ndarray
    :return new_img: numpy.ndarray
    """
    new_img = _image_subtract(gray_img1, gray_img2)

    _debug(visual=new_img,
           filename=os.path.join(params.debug_outdir, str(params.device) + "_subtraction.png"),
           cmap='gray')

    return new_img  # return



import numpy as np
import matplotlib.pyplot as plt
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.cluster.hierarchy import cut_tree
from plantcv.plantcv import params


def _get_empty_count(cur_index_id):
    """Helper function for getting empty count"""
    empty_count = 0
    for i in cur_index_id:
        if i is None:
            empty_count += 1
    return empty_count


def _get_empty_indicies(cur_index, cur_plms_copy):
    """Helper function for getting empty indicies"""
    empty_index = []
    for i, v in zip(cur_index, cur_plms_copy.iloc[cur_index, 0].values):
        if v is None:
            empty_index.append(i)
    return empty_index


def _get_unique_ids(cur_index_id):
    """Helper function for getting id's"""
    unique_ids = []
    for id_ in cur_index_id:
        if id_ is not None and id_ not in unique_ids:
            unique_ids.append(id_)
    return unique_ids


def _get_rogues(cur_plms_copy):
    """Helper function for getting rogues"""
    rogues = []
    for i, x in enumerate(cur_plms_copy['group'].values):
        if x is None:
            rogues.append(i)
    return rogues


def _generate_labelnames(plmnames, grpnames):
    """Helper function for generating label names"""
    labelnames = []
    for li in range(0, len(plmnames)):
        labelname = f'{plmnames[li]} ({grpnames[li]})'
        labelnames.append(labelname)
    return labelnames


def _pair_unnassigned(unique_ids, cur_index, cur_index_id, cur_plms_copy, empty_count, sanity_check_pos):
    """Helper function for pairing unassigned"""
    for uid in unique_ids:
        # If only one plm assigned a name in current cluster and a second unnamed plm exists
        # transfer ID over to create a pair
        if np.count_nonzero(np.array(cur_index_id) == uid) < 2 and empty_count == 1:
            # Store boolean positions for plms with IDs matching current id out of current cluster
            match_ids = [i for i, x in enumerate(cur_plms_copy.iloc[cur_index, 0].values == uid) if x]
            # Store boolean positions for plms which are unnamed out of current cluster
            null_ids = []
            for i, x in enumerate(cur_plms_copy.iloc[cur_index, 0].values):
                if x is None:
                    null_ids.append(i)
            # If exactly 1 matching ID and 1 null ID (i.e. 2 plms total)
            # continue to pass ID name to the unnamed plm
            if len(match_ids) + len(null_ids) == 2:
                # Sanity check! Pairs must be on different days
                pair_names = cur_plms_copy.iloc[[cur_index[i] for i in match_ids + null_ids], 1].values
                if pair_names[0].split('_')[sanity_check_pos] != pair_names[1].split('_')[sanity_check_pos]:
                    # Transfer identities to the unnamed plm
                    cur_plms_copy.iloc[[cur_index[i] for i in null_ids], 0] = uid


def constella(cur_plms, pc_starscape, group_iter, outfile_prefix):
    """
    Group pseudo-landmarks into homology groupings

    Inputs:
    cur_plms       = A pandas array of plm multivariate space representing capturing two adjacent frames in a
                     time series or otherwise analogous dataset in order to enable homology assignments
    pc_starscape   = PCA results from starscape
    group_iter     = Group ID counter
    outfile_prefix = User defined file path and prefix name for PCA output graphics

    :param cur_plms: pandas.core.frame.DataFrame
    :param pc_starscape: pandas.core.frame.DataFrame
    :param group_iter: int
    :param outfile_prefix: str
    """
    # Copy dataframe to avoid modifying the input dataframe
    cur_plms_copy = cur_plms.copy(deep=True)

    sanity_check_pos = 2  # Needs to point at days in image identifier!

    singleton_no = pc_starscape.shape[0]

    if params.debug is not None:
        print(f'{singleton_no} plms to group')

    plm_links = linkage(pc_starscape.loc[:, pc_starscape.columns[2:len(pc_starscape.columns)]].values, 'ward')

    # For n-1 to 2 leaves on the current hierarchical cluster dendrogram...
    for c in np.arange(singleton_no - 1, 2, -1):
        # Extract current number of clusters for the agglomeration step
        cutree = cut_tree(plm_links, n_clusters=c)
        # Generate a list of all current clusters identified
        group_list = np.unique(cutree)

        # For the current cluster being queried...
        for g in group_list:
            # Create list of current clusters row indices in pandas dataframe
            cur_index = [i for i, x in enumerate(cutree == g) if x]
            # Create list of current clusters present group identity assignments
            cur_index_id = np.array(cur_plms_copy.iloc[cur_index, 0])
            # Are any of the plms in the current cluster unnamed, how many?
            empty_count = _get_empty_count(cur_index_id)
            empty_index = _get_empty_indicies(cur_index, cur_plms_copy)
            # Are any of the plms in the current cluster already assigned an identity, what are those identities?
            unique_ids = _get_unique_ids(cur_index_id)
            # If cluster is two unnamed plms exactly, assign this group their own identity as a pair
            if empty_count == 2:
                pair_names = cur_plms_copy.iloc[empty_index, 1].values
                # Sanity check! Pairs must be on different days
                if pair_names[0].split('_')[sanity_check_pos] != pair_names[1].split('_')[sanity_check_pos]:
                    cur_plms_copy.iloc[empty_index, 0] = group_iter
                    group_iter = group_iter + 1
                else:
                    cur_plms_copy.iloc[empty_index[0], 0] = group_iter
                    cur_plms_copy.iloc[empty_index[1], 0] = group_iter + 1
                    group_iter = group_iter + 2
            # If cluster is one unnamed plm and one plm with an identity, assign the unnamed plm the identity of the
            _pair_unnassigned(unique_ids, cur_index, cur_index_id, cur_plms_copy, empty_count, sanity_check_pos)
    rogues = _get_rogues(cur_plms_copy)
    for rogue in rogues:
        cur_plms_copy.iloc[[rogue], 0] = group_iter
        group_iter = group_iter + 1

    grpnames = cur_plms_copy.loc[:, ['group']].values
    plmnames = cur_plms_copy.loc[:, ['plmname']].values

    labelnames = _generate_labelnames(plmnames, grpnames)

    if params.debug is not None:
        plt.figure()
        plt.title('')
        plt.xlabel('')
        plt.ylabel('')
        dendrogram(plm_links, color_threshold=100, orientation="left", leaf_font_size=10, labels=np.array(labelnames))
        plt.tight_layout()

        if params.debug == "print":
            plt.savefig(outfile_prefix + '_plmHCA.png')
            plt.close()
        elif params.debug == "plot":
            plt.show()

    return cur_plms_copy, group_iter



"""Print image to file."""
import cv2
import numpy
import matplotlib
from xarray.core.dataarray import DataArray
from plantcv.plantcv.classes import PSII_data
from plantcv.plantcv import params
from plantcv.plantcv import fatal_error
from plantcv.plantcv._show_dataarray import _show_dataarray
from altair.vegalite.v5.api import FacetChart, LayerChart, Chart


def print_image(img, filename, **kwargs):
    """
    Save image to file.

    Inputs:
    img      = image object
    filename = name of file to save image to
    kwargs   = key-value arguments to xarray.plot method

    :param img: numpy.ndarray, matplotlib.figure.Figure, ggplot, xarray.core.dataarray.DataArray
    :param filename: string
    :param kwargs: dict
    :return:
    """
    # Print numpy array type images
    if isinstance(img, numpy.ndarray):
        cv2.imwrite(filename, img)

    # Print matplotlib type images
    elif isinstance(img, matplotlib.figure.Figure):
        img.savefig(filename, dpi=params.dpi)

    # Print altair type images
    elif isinstance(img, (FacetChart, LayerChart, Chart)):
        img.save(filename, ppi=params.dpi)

    elif isinstance(img, DataArray):
        fig_handle = _show_dataarray(img, **kwargs)
        # fig_handle comes back as a tuple if xarray makes a histogram
        # fig_handle comes back as a list len 1 containing matplotlib.lines.Line2D if xarray makes a line plot
        # will this ever happen? I think _show_dataarray and xarray will fail first
        fig_handle.fig.savefig(filename, dpi=params.dpi)

    elif isinstance(img, PSII_data):
        fatal_error("You need to provide an underlying DataArray.")

    else:
        fatal_error(f"Error writing file {filename}: input img is {str(type(img))}, not a numpy.ndarray, "
                    "matplotlib.figure, plotnine.ggplot, or xarray.core.dataarray.DataArray and cannot get "
                    "saved out with print_image.")



from plantcv.plantcv.roi.roi_methods import circle
from plantcv.plantcv.roi.roi_methods import ellipse
from plantcv.plantcv.roi.roi_methods import from_binary_image
from plantcv.plantcv.roi.roi_methods import rectangle
from plantcv.plantcv.roi.roi_methods import auto_grid
from plantcv.plantcv.roi.roi_methods import multi
from plantcv.plantcv.roi.roi_methods import multi_rect
from plantcv.plantcv.roi.roi_methods import custom
from plantcv.plantcv.roi.roi_methods import filter
from plantcv.plantcv.roi.roi_methods import auto_wells
from plantcv.plantcv.roi.roi2mask import roi2mask
from plantcv.plantcv.roi.quick_filter import quick_filter

__all__ = ["circle", "ellipse", "from_binary_image", "rectangle", "auto_grid",
           "multi", "multi_rect", "custom", "auto_wells",
           "filter", "roi2mask", "quick_filter"]



"""Analyzes spectral index values of objects in an image."""
import os
import numpy as np
from plantcv.plantcv import outputs, params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import fatal_error, warn
from plantcv.plantcv.visualize import histogram
from plantcv.plantcv._helpers import _iterate_analysis


def spectral_index(index_img, labeled_mask, n_labels=1, bins=100, min_bin=0, max_bin=1, label=None):
    """Analyze spectral index values of objects in an image.

    Inputs:
    index_img    = Index image data (PlantCV Spectral_data object)
    labeled_mask = Labeled mask of objects (32-bit).
    n_labels     = Total number expected individual objects (default = 1).
    bins         = Number of histogram bins (default = 100)
    min_bin      = Minimum bin value (default = 0). "auto" will use the minimum value of the index image.
    max_bin      = Maximum bin value (default = 1). "auto" will use the maximum value of the index image.
    label        = Optional label parameter, modifies the variable name of
                   observations recorded (default = pcv.params.sample_label).

    Returns:
    index_hist = Spectral index histogram plot

    :param index_img: plantcv.plantcv.Spectral_data
    :param labeled_mask: numpy.ndarray
    :param n_labels: int
    :param bins: int
    :param min_bin: float
    :param max_bin: float
    :param label: str
    :return index_hist: altair.vegalite.v5.api.FacetChart
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    _ = _iterate_analysis(img=index_img, labeled_mask=labeled_mask, n_labels=n_labels, label=label, function=_analyze_index,
                          **{"bins": bins, "min_bin": min_bin, "max_bin": max_bin})
    index_hist = outputs.plot_dists(variable=f"index_frequencies_{index_img.array_type}")
    _debug(visual=index_hist, filename=os.path.join(params.debug_outdir, str(params.device) + "_index_hist.png"))
    return index_hist


def _analyze_index(img, mask, bins=100, min_bin=0, max_bin=1, label=None):
    """This extracts the hyperspectral index statistics and writes the values  as observations out to
       the Outputs class.

    Inputs:
    img          = Instance of the Spectral_data class, usually the output from pcv.hyperspectral.extract_index
    mask         = Binary mask made from selected contours
    bins         = optional, number of classes to divide spectrum into
    min_bin      = optional, minimum bin value ("auto" or user input minimum value)
    max_bin      = optional, maximum bin value ("auto" or user input maximum value)
    label        = optional label parameter, modifies the variable name of observations recorded

    :param index_array: plantcv.plantcv.Spectral_data
    :param mask: numpy.ndarray
    :param bins: int
    :param max_bin: float, str
    :param min_bin: float, str
    :param label: str
    :return img: plantcv.plantcv.Spectral_data
    """
    debug = params.debug
    params.debug = None

    if len(np.shape(mask)) > 2 or len(np.unique(mask)) > 2:
        fatal_error("Mask should be a binary image of 0 and nonzero values.")

    if len(np.shape(img.array_data)) > 2:
        fatal_error("index_array data should be a grayscale image.")

    # Mask data and collect statistics about pixels within the masked image
    masked_array = img.array_data[np.where(mask > 0)]
    masked_array = masked_array[np.isfinite(masked_array)]

    index_mean = np.nanmean(masked_array)
    index_median = np.nanmedian(masked_array)
    index_std = np.nanstd(masked_array)

    # Set starting point and max bin values
    maxval = max_bin
    b = min_bin
    observed_max = 0
    observed_min = 0

    # Calculate observed min and max pixel values of the masked array
    if masked_array.any():
        observed_max = np.nanmax(masked_array)
        observed_min = np.nanmin(masked_array)

    # Auto calculate max_bin if set
    if type(max_bin) is str and (max_bin.upper() == "AUTO"):
        maxval = float(round(observed_max, 8))  # Auto bins will detect maxval to use for calculating labels/bins
    if type(min_bin) is str and (min_bin.upper() == "AUTO"):
        b = float(round(observed_min, 8))  # If bin_min is auto then overwrite starting value

    # Print a warning if observed min/max outside user defined range
    if observed_max > maxval or observed_min < b:
        warn(f"The observed range of pixel values in your masked index provided is [{str(observed_min)}"
             f", {str(observed_max)}] but the user defined range of bins for pixel frequencies is [{str(b)}, "
             f"{str(maxval)}]. Adjust min_bin and max_bin in order to avoid cutting off data being collected.")

    # Calculate histogram
    bin_labels, hist_percent = [], []
    if mask.any():
        _, hist_data = histogram(img.array_data, mask=mask, bins=bins, lower_bound=b, upper_bound=maxval,
                                 hist_data=True)
        bin_labels, hist_percent = hist_data["pixel intensity"].tolist(), hist_data["proportion of pixels (%)"].tolist()

    # Restore user debug setting
    params.debug = debug

    outputs.add_observation(sample=label, variable=f"mean_{img.array_type}", trait=f"Average {img.array_type} reflectance",
                            method="plantcv.plantcv.analyze.spectral_index", scale="reflectance", datatype=float,
                            value=float(index_mean), label="none")

    outputs.add_observation(sample=label, variable=f"med_{img.array_type}", trait=f"Median {img.array_type} reflectance",
                            method="plantcv.plantcv.analyze.spectral_index", scale="reflectance", datatype=float,
                            value=float(index_median), label="none")

    outputs.add_observation(sample=label, variable=f"std_{img.array_type}",
                            trait=f"Standard deviation {img.array_type} reflectance",
                            method="plantcv.plantcv.analyze.spectral_index", scale="reflectance", datatype=float,
                            value=float(index_std), label="none")

    outputs.add_observation(sample=label, variable=f"index_frequencies_{img.array_type}",
                            trait="index frequencies", method="plantcv.plantcv.analyze.spectral_index", scale="frequency",
                            datatype=list, value=hist_percent, label=bin_labels)
    return img



import pandas as pd
import numpy as np
from plantcv.plantcv import params


def constellaqc(denovo_groups, annotated_groups):
    """
    Compare de novo annotations of Constella to known standards in order to estimate error rates

    Inputs:
    denovo_groups    = A pandas array representing homology groups predicted by Constella for plms
    annotated_groups = A pandas array representing the true biological identities of plms

    :param denovo_groups: pandas.core.frame.DataFrame
    :param annotated_groups: pandas.core.frame.DataFrame

    """
    known_feat = np.unique(annotated_groups.loc[:, 'group'])
    pred_group = np.unique(denovo_groups.loc[:, 'group'])

    scores = []

    for anno in known_feat:
        anno_group_calls = denovo_groups.loc[annotated_groups.loc[:, 'group'] == anno, 'group'].values
        score_row = []
        for denovo in pred_group:
            score_row.append(np.sum(anno_group_calls == denovo))
        scores.append(score_row)

    scores = pd.DataFrame(scores, index=known_feat, columns=pred_group)

    if params.debug is not None:
        print('Known Feature-Predicted Group Scoring Matrix:\n')
        print(scores)

    anno_sum = []
    anno_no = []
    anno_error = []
    ni = []

    for anno in known_feat:
        anno_sum.append(np.sum(scores.loc[anno, :].values))
        anno_no.append(np.sum(scores.loc[anno, :].values != 0))
        anno_error.append(np.sum(scores.loc[anno, :].values != 0) - 1)
        ni.append(1)
    pred_sum = []
    pred_no = []
    pred_error = []
    nj = []

    for denovo in pred_group:
        pred_sum.append(np.sum(scores.loc[:, denovo].values))
        pred_no.append(np.sum(scores.loc[:, denovo].values != 0))
        pred_error.append(np.sum(scores.loc[:, denovo].values != 0) - 1)
        nj.append(1)

    anno_valid = np.array(anno_sum) - ni - np.array(anno_error)

    v_sum = np.sum(anno_valid)
    s_sum = np.sum(anno_error)
    c_sum = np.sum(pred_error)
    total = v_sum + s_sum + c_sum

    print('\n\nValid Call Rate:     ', round(100 * (v_sum / total), 2), '%')
    print('Splitting Call Rate: ', round(100 * (s_sum / total), 2), '%')
    print('Clumping Call Rate:  ', round(100 * (c_sum / total), 2), '%')



# RGB -> HSV -> Gray

import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params
from plantcv.plantcv._helpers import _rgb2hsv


def rgb2gray_hsv(rgb_img, channel):
    """Convert image from RGB colorspace to HSV colorspace. Returns the specified subchannel as a gray image.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image data
    channel : str
        color subchannel (h = hue, s = saturation, v = value/intensity/brightness)

    Returns
    -------
    numpy.ndarray
        grayscale image from one HSV color channel
    """
    # Convert RGB to HSV and return the specified subchannel as a gray image
    gray_img = _rgb2hsv(rgb_img=rgb_img, channel=channel)

    # The allowable channel inputs are h, s or v
    names = {"h": "hue", "s": "saturation", "v": "value"}

    _debug(visual=gray_img,
           filename=os.path.join(params.debug_outdir, f"{params.device}_hsv_{names[channel.lower()]}.png"), cmap='gray')

    return gray_img



from importlib.metadata import version

# Auto versioning
__version__ = version("plantcv")

from plantcv.plantcv.fatal_error import fatal_error
from plantcv.plantcv.classes import Params
from plantcv.plantcv.classes import Outputs
from plantcv.plantcv.classes import Spectral_data
from plantcv.plantcv.classes import PSII_data
from plantcv.plantcv.classes import Points
from plantcv.plantcv.classes import Objects

# Initialize an instance of the Params and Outputs class with default values
# params and outputs are available when plantcv is imported
params = Params()
outputs = Outputs()

from plantcv.plantcv.deprecation_warning import deprecation_warning
from plantcv.plantcv.warn import warn
from plantcv.plantcv.print_image import print_image
from plantcv.plantcv.plot_image import plot_image
from plantcv.plantcv.color_palette import color_palette
from plantcv.plantcv.rgb2gray import rgb2gray
from plantcv.plantcv.rgb2gray_hsv import rgb2gray_hsv
from plantcv.plantcv.rgb2gray_lab import rgb2gray_lab
from plantcv.plantcv.rgb2gray_cmyk import rgb2gray_cmyk
from plantcv.plantcv.gaussian_blur import gaussian_blur
from plantcv.plantcv import transform
from plantcv.plantcv import hyperspectral
from plantcv.plantcv import spectral_index
from plantcv.plantcv.apply_mask import apply_mask
from plantcv.plantcv.readimage import readimage
from plantcv.plantcv.readbayer import readbayer
from plantcv.plantcv.laplace_filter import laplace_filter
from plantcv.plantcv.sobel_filter import sobel_filter
from plantcv.plantcv.scharr_filter import scharr_filter
from plantcv.plantcv.hist_equalization import hist_equalization
from plantcv.plantcv.image_add import image_add
from plantcv.plantcv.image_fusion import image_fusion
from plantcv.plantcv.image_subtract import image_subtract
from plantcv.plantcv.erode import erode
from plantcv.plantcv.dilate import dilate
from plantcv.plantcv.watershed import watershed_segmentation
from plantcv.plantcv.median_blur import median_blur
from plantcv.plantcv.fill import fill
from plantcv.plantcv.invert import invert
from plantcv.plantcv.logical_and import logical_and
from plantcv.plantcv.logical_or import logical_or
from plantcv.plantcv.logical_xor import logical_xor
from plantcv.plantcv.within_frame import within_frame
from plantcv.plantcv.flip import flip
from plantcv.plantcv.crop_position_mask import crop_position_mask
from plantcv.plantcv.report_size_marker_area import report_size_marker_area
from plantcv.plantcv.white_balance import white_balance
from plantcv.plantcv.shift_img import shift_img
from plantcv.plantcv.output_mask_ori_img import output_mask
from plantcv.plantcv.auto_crop import auto_crop
from plantcv.plantcv.background_subtraction import background_subtraction
from plantcv.plantcv.naive_bayes_classifier import naive_bayes_classifier
from plantcv.plantcv import homology
from plantcv.plantcv.distance_transform import distance_transform
from plantcv.plantcv.canny_edge_detect import canny_edge_detect
from plantcv.plantcv.opening import opening
from plantcv.plantcv.closing import closing
from plantcv.plantcv import roi
from plantcv.plantcv import threshold
from plantcv.plantcv import visualize
from plantcv.plantcv import morphology
from plantcv.plantcv.fill_holes import fill_holes
from plantcv.plantcv.get_kernel import get_kernel
from plantcv.plantcv.crop import crop
from plantcv.plantcv.stdev_filter import stdev_filter
from plantcv.plantcv.spatial_clustering import spatial_clustering
from plantcv.plantcv import photosynthesis
from plantcv.plantcv import annotate
from plantcv.plantcv import io
from plantcv.plantcv.segment_image_series import segment_image_series
from plantcv.plantcv.create_labels import create_labels
from plantcv.plantcv.floodfill import floodfill
from plantcv.plantcv import analyze
from plantcv.plantcv import filters
from plantcv.plantcv.kmeans_classifier import predict_kmeans
from plantcv.plantcv.kmeans_classifier import mask_kmeans
from plantcv.plantcv import qc
# add new functions to end of lists

__all__ = [
    "fatal_error",
    "Params",
    "Outputs",
    "Spectral_data",
    'PSII_data',
    'Points',
    "Objects",
    "deprecation_warning",
    "warn",
    "print_image",
    "plot_image",
    "color_palette",
    "rgb2gray",
    "rgb2gray_hsv",
    "rgb2gray_lab",
    "rgb2gray_cmyk",
    "gaussian_blur",
    "transform",
    "hyperspectral",
    "spectral_index",
    "apply_mask",
    "readimage",
    "readbayer",
    "laplace_filter",
    "sobel_filter",
    "scharr_filter",
    "hist_equalization",
    "image_add",
    "image_fusion",
    "image_subtract",
    "erode",
    "dilate",
    "watershed_segmentation",
    "median_blur",
    "fill",
    "invert",
    "logical_and",
    "logical_or",
    "logical_xor",
    "within_frame",
    "flip",
    "crop_position_mask",
    "report_size_marker_area",
    "white_balance",
    "shift_img",
    "output_mask",
    "auto_crop",
    "background_subtraction",
    "naive_bayes_classifier",
    "distance_transform",
    "canny_edge_detect",
    "opening",
    "closing",
    "roi",
    "threshold",
    "visualize",
    "morphology",
    "fill_holes",
    "get_kernel",
    "crop",
    "stdev_filter",
    "spatial_clustering",
    "photosynthesis",
    "homology",
    "annotate",
    "io",
    "segment_image_series",
    "create_labels",
    "analyze",
    "floodfill",
    "filters",
    "predict_kmeans",
    "mask_kmeans",
    "qc"
]



# Sort segments

import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv._helpers import _find_tips, _logical_operation, _dilate
from plantcv.plantcv._debug import _debug


def segment_sort(skel_img, objects, mask=None, first_stem=True):
    """Sort segments from a skeletonized image into two categories: leaf objects and other objects.

    Inputs:
    skel_img          = Skeletonized image
    objects           = List of contours
    mask              = (Optional) binary mask for debugging. If provided, debug image will be overlaid on the mask.
    first_stem        = (Optional) if True, then the first (bottom) segment always gets classified as stem

    Returns:
    labeled_img       = Segmented debugging image with lengths labeled
    secondary_objects = List of secondary segments (leaf)
    primary_objects   = List of primary objects (stem)

    :param skel_img: numpy.ndarray
    :param objects: list
    :param mask: numpy.ndarray
    :param first_stem: bool
    :return secondary_objects: list
    :return other_objects: list
    """
    secondary_objects = []
    primary_objects = []

    if mask is None:
        labeled_img = np.zeros(skel_img.shape[:2], np.uint8)
    else:
        labeled_img = mask.copy()

    tips_img, _, _ = _find_tips(skel_img)
    tips_img = _dilate(tips_img, 3, 1)

    # Loop through segment contours
    for i, cnt in enumerate(objects):
        segment_plot = np.zeros(skel_img.shape[:2], np.uint8)
        cv2.drawContours(segment_plot, objects, i, 255, 1, lineType=8)
        overlap_img = _logical_operation(segment_plot, tips_img, "and")

        # The first contour is the base, and while it contains a tip, it isn't a leaf
        if i == 0 and first_stem:
            primary_objects.append(cnt)

        # Sort segments
        else:

            if np.sum(overlap_img) > 0:
                secondary_objects.append(cnt)
            else:
                primary_objects.append(cnt)

    # Plot segments where green segments are leaf objects and fuschia are other objects
    labeled_img = cv2.cvtColor(labeled_img, cv2.COLOR_GRAY2RGB)
    for i, cnt in enumerate(primary_objects):
        cv2.drawContours(labeled_img, primary_objects, i, (255, 0, 255), params.line_thickness, lineType=8)
    for i, cnt in enumerate(secondary_objects):
        cv2.drawContours(labeled_img, secondary_objects, i, (0, 255, 0), params.line_thickness, lineType=8)

    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_sorted_segments.png"))

    return secondary_objects, primary_objects



# Join images (OR)

import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params
from plantcv.plantcv._helpers import _logical_operation


def logical_or(bin_img1, bin_img2):
    """Join two images using the bitwise OR operator.

    Parameters
    ----------
    bin_img1 : numpy.ndarray
        First binary image.
    bin_img2 : numpy.ndarray
        Second binary image.

    Returns
    -------
    numpy.ndarray
        Joined binary image.
    """
    merged = _logical_operation(bin_img1, bin_img2, 'or')

    _debug(visual=merged, filename=os.path.join(params.debug_outdir, f'{params.device}_or_joined.png'), cmap='gray')

    return merged



"""Fluorescence Analysis (Fv/Fm parameter)."""
import os
import numpy as np
import pandas as pd
import xarray as xr
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params, outputs, fatal_error
from plantcv.plantcv.photosynthesis import reassign_frame_labels


def yii(ps_da, labeled_mask, n_labels=1, auto_fm=False, measurement_labels=None, label=None):
    """
    Calculate and analyze PSII efficiency estimates from fluorescence image data.

    Inputs:
    ps_da               = Photosynthesis xarray DataArray (either ojip_dark, ojip_light, pam_dark, or pam_light)
    labeled_mask        = Labeled mask of objects (32-bit).
    n_labels            = Total number expected individual objects (default = 1).
    auto_fm             = Automatically calculate the frame with maximum fluorescence per label, otherwise
                          use a fixed frame for all labels (default = False).
    measurement_labels  = labels for each measurement, modifies the variable name of observations recorded
    label               = optional label parameter, modifies the variable name of observations recorded

    Returns:
    yii_global          = DataArray of efficiency estimate values
    yii_chart           = Histograms of efficiency estimate

    :param ps_da: xarray.core.dataarray.DataArray
    :param labeled_mask: numpy.ndarray
    :param n_labels: int
    :param auto_fm: bool
    :param measurement_labels: list
    :param label: str
    :return yii_global: xarray.core.dataarray.DataArray
    :return yii_chart: altair.vegalite.v4.api.FacetChart
    """
    # Set labels
    labels = _set_labels(label, n_labels)

    # Validate that the input mask has the same 2D shape as the input DataArray
    if labeled_mask.shape != ps_da.shape[:2]:
        fatal_error(f"Mask needs to have shape {ps_da.shape[:2]}")

    # Validate that the input measurement_labels is the same length as the number of measurements in the DataArray
    if (measurement_labels is not None) and (len(measurement_labels) != ps_da.coords['measurement'].shape[0]):
        fatal_error('measurement_labels must be the same length as the number of measurements in the DataArray')

    # The name of the DataArray
    var = ps_da.name.lower()

    # Validate that var is a supported type
    if var not in ['ojip_dark', 'ojip_light', 'pam_dark', 'pam_light']:
        fatal_error(f"Unsupported DataArray type: {var}")

    # Make an zeroed array of the same shape as the input DataArray
    yii_global = xr.zeros_like(ps_da, dtype=float)
    # Drop the frame_label coordinate
    yii_global = yii_global[:, :, 0, :].drop_vars('frame_label')

    # Make a copy of the labeled mask
    mask_copy = np.copy(labeled_mask)

    # If the labeled mask is a binary mask with values 0 and 255, convert to 0 and 1
    if len(np.unique(mask_copy)) == 2 and np.max(mask_copy) == 255:
        mask_copy = np.where(mask_copy == 255, 1, 0).astype(np.uint8)

    # Iterate over the label values 1 to n_labels
    for i in range(1, n_labels + 1):
        # Create a binary submask for each label
        submask = np.where(mask_copy == i, 255, 0).astype(np.uint8)

        # Expand the submask to the same shape as the input DataArray
        submask = submask[..., None, None]

        # If auto_fm is True, reassign frame labels to choose the best Fm or Fm' for each labeled region
        if auto_fm:
            ps_da = reassign_frame_labels(ps_da=ps_da, mask=submask.squeeze().squeeze())

        # Mask the input DataArray with the submask
        yii_masked = ps_da.astype('float').where(submask > 0, other=np.nan)

        # Dark-adapted datasets (Fv/Fm)
        if var in ['ojip_dark', 'pam_dark']:
            # Calculate Fv/Fm
            yii_lbl = (yii_masked.sel(frame_label='Fm') - yii_masked.sel(frame_label='F0')) / yii_masked.sel(frame_label='Fm')

        # Light-adapted datasets (Fq'/Fm')
        if var in ['ojip_light', 'pam_light']:
            # Calculate Fq'/Fm'
            yii_lbl = yii_masked.groupby('measurement', squeeze=False).map(_calc_yii)

        # Drop the frame_label coordinate
        yii_lbl = yii_lbl.drop_vars('frame_label')
        # Fill NaN values with 0 so that we can add DataArrays together
        yii_lbl = yii_lbl.fillna(0)
        # Add the Fv/Fm values for this label to the yii DataArray
        yii_global = yii_global + yii_lbl

        # Record observations for each labeled region
        _add_observations(yii_da=yii_lbl, measurements=ps_da.measurement.values, label=f"{labels[i - 1]}_{i}",
                          measurement_labels=measurement_labels)

    # Convert the labeled mask to a binary mask
    bin_mask = np.where(labeled_mask > 0, 255, 0)

    # Expand the binary mask to the same shape as the YII DataArray
    bin_mask = bin_mask[..., None]

    # Set the background values to NaN
    yii_global = yii_global.where(bin_mask > 0, other=np.nan)

    # drop coords identifying frames if they exist
    res = [i for i in list(yii_global.coords) if 'frame' in i]
    yii_global = yii_global.drop_vars(res)  # does not fail if res is []

    # Create a ridgeline plot of the YII values
    yii_chart = _ridgeline_plots(measurements=ps_da.measurement.values, measurement_labels=measurement_labels)

    # Create a pseudocolor image of the YII values
    _debug(visual=yii_global,
           filename=os.path.join(params.debug_outdir, str(params.device) + "_YII_dataarray.png"),
           robust=True,
           col='measurement',
           col_wrap=int(np.ceil(yii_global.measurement.size / 4)),
           vmin=0, vmax=1)

    return yii_global.squeeze(), yii_chart


def _set_labels(label, n_labels):
    """Create list of labels."""
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label
    # Set labels to label
    labels = label
    # If label is a string, make a list of labels
    if isinstance(label, str):
        labels = [label] * n_labels
    # If the length of the labels list is not equal to the number of labels, raise an error
    if len(labels) != n_labels:
        fatal_error(f"Number of labels ({len(labels)}) does not match number of objects ({n_labels})")
    return labels


def _create_histogram(yii_img, mlabel):
    """
    Compute histogram of YII.

    Inputs:
    yii_img     = numpy array of yii
    mlabel      = measurement label
    obs         = PlantCV observations used to retrieve statistics

    Returns:
    hist_df    = Histogram of efficiency estimate
    yii_mode   = DataArray of efficiency estimate values

    :param yii_img: numpy.ndarray
    :param mlabel: str
    :param obs: dict
    :return hist_df: pandas.DataFrame
    :return yii_mode: float
    """
    # Calculate the histogram of Fv/Fm, Fv'/Fm', or Fq'/Fm' non-zero values
    yii_hist, yii_bins = np.histogram(yii_img[np.where(yii_img > 0)], 100, range=(0, 1))
    # yii_bins is a bins + 1 length list of bin endpoints, so we need to calculate bin midpoints so that
    # the we have a one-to-one list of x (YII) and y (frequency) values.
    # To do this we add half the bin width to each lower bin edge x-value
    # midpoints = yii_bins[:-1] + 0.5 * np.diff(yii_bins)

    # Calculate which non-zero bin has the maximum Fv/Fm value
    yii_mode = yii_bins[np.argmax(yii_hist)]

    # Convert the histogram pixel counts to proportional frequencies
    yii_percent = (yii_hist / float(np.sum(yii_hist))) * 100

    # Create a dataframe for the histogram
    hist_df = pd.DataFrame({'proportion of pixels (%)': yii_percent, mlabel: yii_bins[:-1]})

    return hist_df, yii_mode


def _add_observations(yii_da, measurements, measurement_labels, label):
    """Add observations for each labeled region."""
    # compute observations to store in Outputs, per labeled region
    yii_mean = yii_da.where(yii_da > 0).groupby('measurement').mean(['x', 'y']).values
    yii_median = yii_da.where(yii_da > 0).groupby('measurement').median(['x', 'y']).values
    yii_max = yii_da.where(yii_da > 0).groupby('measurement').max(['x', 'y']).values

    # Create variables to label traits based on measurement label in data array
    for n, mlabel in enumerate(measurements):
        if measurement_labels is not None:
            mlabel = measurement_labels[n]

        # mean value
        outputs.add_observation(sample=label, variable=f"yii_mean_{mlabel}", trait="mean yii value",
                                method='plantcv.plantcv.analyze.yii', scale='none', datatype=float,
                                value=float(yii_mean[n]), label='none')
        # median value
        outputs.add_observation(sample=label, variable=f"yii_median_{mlabel}", trait="median yii value",
                                method='plantcv.plantcv.analyze.yii', scale='none', datatype=float,
                                value=float(yii_median[n]), label='none')
        # max value
        outputs.add_observation(sample=label, variable=f"yii_max_{mlabel}", trait="peak yii value",
                                method='plantcv.plantcv.analyze.yii', scale='none', datatype=float,
                                value=float(yii_max[n]), label='none')

        hist_df, yii_mode = _create_histogram(yii_da.isel({'measurement': n}).values, mlabel)

        # mode value
        outputs.add_observation(sample=label, variable=f"yii_mode_{mlabel}", trait="mode yii value",
                                method='plantcv.plantcv.analyze.yii', scale='none', datatype=float,
                                value=float(yii_mode), label='none')
        # hist frequencies
        outputs.add_observation(sample=label, variable=f"yii_hist_{mlabel}", trait="yii frequencies",
                                method='plantcv.plantcv.analyze.yii', scale='none', datatype=list,
                                value=hist_df['proportion of pixels (%)'].values.tolist(),
                                label=np.around(hist_df[mlabel].values.tolist(), decimals=2).tolist())


def _calc_yii(da):
    """Apply the Fq'/Fm' calculation to the DataArray."""
    return (da.sel(frame_label='Fmp') - da.sel(frame_label='Fp')) / da.sel(frame_label='Fmp')


def _ridgeline_plots(measurements, measurement_labels):
    """Create ridgeline plots of YII values."""
    yii_chart = None
    for i, mlabel in enumerate(measurements):
        if measurement_labels is not None:
            mlabel = measurement_labels[i]
        yii_chart = outputs.plot_dists(variable=f"yii_hist_{mlabel}")
        _debug(visual=yii_chart, filename=os.path.join(params.debug_outdir, str(params.device) + '_yii_hist.png'))
    return yii_chart



# Create time-lapse videos with input directory of images

import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import fatal_error
from plantcv.plantcv.transform import resize
from plantcv.plantcv import warn


def time_lapse_video(img_list, out_filename='./time_lapse_video.mp4', fps=29.97, display=True):
    """Generate time-lapse video given a list of paths to the images

    Inputs:
    img_list       = the desired list of paths to the images to create the video
    out_filename   = name of file to save the generated video to
    fps            = frame rate (frames per second)
    display        = if True (default), displays the path to the generated video

    Outputs:
    frame_size     = the frame size of the generated video

    :param img_list: list
    :param fps: float
    :param out_filename: string
    :param display: boolean
    :return frame_size: tuple
    """
    params.debug = None

    if len(img_list) <= 0:
        fatal_error("Image list is empty")

    imgs = []
    list_r = []
    list_c = []
    for file in img_list:
        img = cv2.imread(file)
        if img is None:
            fatal_error(f"Unable to read {file}")
        list_r.append(img.shape[0])
        list_c.append(img.shape[1])
        imgs.append(img)
    max_c, max_r = np.max(list_c), np.max(list_r)

    # use the largest size of the images as the frame size
    # frame_size = frame_size or (max_c, max_r)
    frame_size = (max_c, max_r)

    if not (len(np.unique(list_r)) == 1 and len(np.unique(list_c)) == 1):
        warn("The sizes of images are not the same, an image resizing (cropping or zero-padding) will be done "
             f"to make all images the same size ({frame_size[0]}x{frame_size[1]}) before creating the video! ")

    out_path, _ = os.path.splitext(out_filename)
    out_filename = out_path + '.mp4'

    fourcc = cv2.VideoWriter_fourcc(*'mp4v')  # Be sure to use lower case
    out = cv2.VideoWriter(out_filename, fourcc, fps, frame_size)

    for img in imgs:
        out.write(resize(img, frame_size, interpolation=None))
    out.release()
    cv2.destroyAllWindows()
    if display is True:
        print(f'Path to generated video: \n{out_filename}')

    return frame_size



# Generate a plm multivariate space for downstream use in homology group assignments

from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from plantcv.plantcv import params


def starscape(cur_plms, group_a, group_b, outfile_prefix):
    """
    Principal Component Analysis on pseudo-landmark data between two timepoints

    Inputs:
    cur_plms        = A pandas array of plm multivariate space representing capturing two adjacent frames in a
                      time series or otherwise analogous dataset in order to enable homology assignments
    group_a         = Name of group A (timepoint 1)
    group_b         = Name of group B (timepoint 2)
    outfile_prefix  = User defined file path and prefix name for PCA output graphics

    Outputs:
    final_df        = Output dataframe
    eigenvals       = PCA eigen values
    loadings        = PCA loadings

    :param cur_plms: pandas.core.frame.DataFrame
    :param group_a: str
    :param group_b: str
    :param outfile_prefix: str
    :return final_df: pandas.core.frame.DataFrame
    :return eigenvals: numpy.ndarray
    :return loadings: pandas.core.frame.DataFrame
    """
    init_comps = 4

    # Store names of our features/dimensions used to generate our PCA
    features = cur_plms.columns[3:len(cur_plms.columns)].tolist()

    # Remove label columns prior to running PCA
    x = cur_plms.loc[:, features].values
    # Rescale dataframe values
    scaler = StandardScaler()
    scaler.fit(x)
    x_scaled = scaler.transform(x)

    # Set initial number of components to fit PC space to
    pca2 = PCA(init_comps)
    # Fit PCA space to rescaled dataframe
    pca2.fit(x_scaled)

    # Store eigenvalues
    eigenvals = pca2.explained_variance_
    # Store variance explained by each component
    var_exp = pca2.explained_variance_ratio_

    if params.debug is not None:
        # Print cumulative variance explained by each component
        print('Eigenvalues: ', eigenvals, '\n\n')
        # Print cumulative variance explained by each component
        print('Var. Explained: ', var_exp, '\n\n')
        # Print cumulative variance explained by each component
        print('Cumul. Var. Explained: ', np.cumsum(pca2.explained_variance_ratio_), '\n\n')

    # Store the number of informative components required to surpass cutoff
    informative_comps = sum(eigenvals > 1)

    print(informative_comps, ' components sufficiently informative')

    # Rerun PCA to extract informative components specifically
    pca = PCA(informative_comps)
    # Fit Second PCA space to rescaled dataframe
    pca.fit(x_scaled)

    principal_components = pca.fit_transform(x)

    principal_df = pd.DataFrame(data=principal_components,
                                columns=["PC" + str(i) for i in range(1, informative_comps + 1)])
    cur_plm_names = pd.DataFrame(cur_plms.loc[:, 'plmname'].values)
    cur_plm_names.columns = ['plmname']
    cur_filenames = pd.DataFrame(cur_plms.loc[:, 'filename'].values)
    cur_filenames.columns = ['filename']

    final_df = pd.concat([cur_plm_names, cur_filenames, principal_df], axis=1)

    loadings = pd.DataFrame(pca.components_.T, columns=["PC" + str(i) for i in range(1, informative_comps + 1)],
                            index=features)

    if params.debug is not None:
        # Generate Screeplot for PCA of plm space
        plt.figure()
        plt.plot(np.arange(1, len(eigenvals) + 1), eigenvals)
        plt.hlines(1, 1, len(eigenvals) + 1, linestyles='dotted', color='r')
        plt.title('Screeplot for Principal Components of plm Space')
        plt.xlabel('Number of Components', fontsize=12)
        plt.ylabel('Eigenvalues', fontsize=12)

        if params.debug == 'print':
            plt.savefig(f"{outfile_prefix}_screeplot.png", dpi=params.dpi)
            plt.close()
        elif params.debug == 'plot':
            # Use non-blocking mode in case the function is run more than once
            plt.show(block=False)

        # Plot principal components
        plt.figure()
        targets = [group_a, group_b]
        colors = ['r', 'b']
        if informative_comps >= 3:
            # 3D plot of first 3 PCA dimensions
            prcomp = plt.axes(projection='3d')
            # Set axis labels
            prcomp.set_xlabel('Principal Component 1')
            prcomp.set_ylabel('Principal Component 2')
            prcomp.set_zlabel('Principal Component 3')

            indices_to_keep = final_df['filename'] == targets[0]
            prcomp.scatter3D(final_df.loc[indices_to_keep, 'PC1'],
                             final_df.loc[indices_to_keep, 'PC2'],
                             final_df.loc[indices_to_keep, 'PC3'], c=colors[0])

            indices_to_keep = final_df['filename'] == targets[1]
            prcomp.scatter3D(final_df.loc[indices_to_keep, 'PC1'],
                             final_df.loc[indices_to_keep, 'PC2'],
                             final_df.loc[indices_to_keep, 'PC3'], c=colors[1])
        elif informative_comps == 2:
            # 2D plot of First 2 PCA dimensions
            indices_to_keep = final_df['filename'] == targets[0]
            plt.scatter(final_df.loc[indices_to_keep, 'PC1'], final_df.loc[indices_to_keep, 'PC2'], c=colors[0])

            indices_to_keep = final_df['filename'] == targets[1]
            plt.scatter(final_df.loc[indices_to_keep, 'PC1'], final_df.loc[indices_to_keep, 'PC2'], c=colors[1])

            plt.xlabel('Principal Component 1')
            plt.ylabel('Principal Component 2')

        if params.debug == "print":
            plt.savefig(f"{outfile_prefix}_PCspace.png", dpi=params.dpi)
            plt.close()
        elif params.debug == "plot":
            plt.show(block=False)

    return final_df, eigenvals, loadings



# Rescale grayscale images to user defined range

import os
import numpy as np
from plantcv.plantcv import fatal_error
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug


def rescale(gray_img, min_value=0, max_value=255):
    """Rescale image.

    Inputs:
    gray_img  = Grayscale image data
    min_value = (optional) new minimum value for range of interest. default = 0
    max_value = (optional) new maximum value for range of interest. default = 255

    Returns:
    rescaled_img = rescaled image

    :param gray_img: numpy.ndarray
    :param min_value: int
    :param max_value: int
    :return c: numpy.ndarray
    """
    if len(np.shape(gray_img)) != 2:
        fatal_error("Image is not grayscale")

    rescaled_img = np.interp(gray_img, (np.nanmin(gray_img), np.nanmax(gray_img)), (min_value, max_value))
    rescaled_img = (rescaled_img).astype('uint8')

    _debug(visual=rescaled_img, filename=os.path.join(params.debug_outdir, str(params.device) + '_rescaled.png'))

    return rescaled_img
